from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, NamedTuple
from unittest import mock

import pytest

from dissect.target.plugins.os.windows.env import (
    EnvironmentVariablePlugin,
    EnvVarDetails,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import RegistryKey

PATHEXTS = [".ext_1", ".ext_2"]


class _MockUser(NamedTuple):
    name: str
    home: str
    sid: str


MockUser = _MockUser(
    "test-user",
    "test-user-home",
    "test-user-sid",
)


TEST_VARIABLES = [
    EnvVarDetails(
        "%foo%",
        (
            "foo1",
            "foo2",
        ),
        "FOO",
    ),
    EnvVarDetails(
        "%path%",
        (
            "path1;;;",
            "path2;;;",
        ),
        "PATH",
    ),
]

TEST_USER_VARIABLES = [
    EnvVarDetails(
        "%path%",
        (
            "{user_sid}/path1;;;",
            "{user_sid}/path2;;;",
        ),
        "PATH",
    ),
]

TEST_USER_ENV_VARS = (
    (
        None,
        (
            ("%foo%", "foo2"),
            ("%path%", "path1;path2;;;"),
        ),
    ),
    (
        MockUser,
        (
            ("%foo%", "foo2"),
            ("%path%", f"{MockUser.sid}/path1;test-user-sid/path2;;;"),
        ),
    ),
)


@pytest.fixture
def env_plugin() -> Iterator[EnvironmentVariablePlugin]:
    mock_target = mock.Mock()

    class _MockValue(NamedTuple):
        value: str

    def registry_value_side_effect(key: RegistryKey, value: str) -> _MockValue:
        return _MockValue(key)

    mock_registry_value = mock.Mock(side_effect=registry_value_side_effect)
    mock_target.registry.value = mock_registry_value

    mock_users = mock.Mock(side_effect=lambda: [MockUser])
    mock_target.users = mock_users

    mock_target.path = "mock"
    mock_target._config.CACHE_DIR = "cache"

    with (
        mock.patch.object(EnvironmentVariablePlugin, "VARIABLES", TEST_VARIABLES),
        mock.patch.object(EnvironmentVariablePlugin, "USER_VARIABLES", TEST_USER_VARIABLES),
    ):
        plugin = EnvironmentVariablePlugin(mock_target)
        plugin._pathext = set(PATHEXTS)

        yield plugin


def test__expand_env() -> None:
    test_value = "%foo%%bAr%/%FoO%foo%bar%bar"
    test_env_vars = OrderedDict([("%FOO%", "bar"), ("%bar%", "foo")])
    expanded_value = EnvironmentVariablePlugin._expand_env(test_value, test_env_vars)
    assert expanded_value == "barfoo/barfoofoobar"


def test__expand_env_vars() -> None:
    test_env_vars = OrderedDict(
        [
            ("%first%", "first"),
            ("%second%", "%first%"),
            ("%third%", "%fourth%"),
            ("%fourth%", "%second%"),
        ]
    )

    result_env_vars = OrderedDict(
        [
            ("%first%", "first"),
            ("%second%", "first"),
            ("%third%", "%fourth%"),
            ("%fourth%", "first"),
        ]
    )

    assert result_env_vars == EnvironmentVariablePlugin._expand_env_vars(test_env_vars)


def test__get_env_vars(env_plugin: EnvironmentVariablePlugin) -> None:
    env_vars = env_plugin._get_env_vars(env_plugin.VARIABLES)

    assert env_vars["%foo%"] == "foo2"
    assert env_vars["%path%"] == "path1;path2;;;"


def test__get_system_env_vars(env_plugin: EnvironmentVariablePlugin) -> None:
    with mock.patch.object(env_plugin, "_get_env_vars"), mock.patch.object(env_plugin, "_expand_env_vars"):
        env_plugin._get_system_env_vars()
        env_plugin._get_env_vars.assert_called_with(env_plugin.VARIABLES)
        env_plugin._expand_env_vars.assert_called_with(env_plugin._get_env_vars.return_value)


@pytest.mark.parametrize(("user", "results"), TEST_USER_ENV_VARS)
def test__get_user_env_vars(
    env_plugin: EnvironmentVariablePlugin, user: MockUser, results: tuple[tuple[str, str]]
) -> None:
    user_sid = None
    if user:
        user_sid = user.sid
    env_vars = env_plugin._get_user_env_vars(user_sid)

    for variable, value in results:
        assert env_vars[variable] == value


@pytest.mark.parametrize(("user", "results"), TEST_USER_ENV_VARS)
def test_expand_env(env_plugin: EnvironmentVariablePlugin, user: MockUser, results: tuple[tuple[str, str]]) -> None:
    path = "mock"

    with mock.patch.object(env_plugin, "_get_user_env_vars"), mock.patch.object(env_plugin, "_expand_env"):
        user_sid = None
        if user:
            user_sid = user.sid
        expanded_path = env_plugin.expand_env(path, user_sid)
        env_plugin._get_user_env_vars.assert_called_with(user_sid)
        env_plugin._expand_env.assert_called_with(path, env_plugin._get_user_env_vars.return_value)
        assert expanded_path == env_plugin._expand_env.return_value


@pytest.mark.parametrize(("user", "results"), TEST_USER_ENV_VARS)
def test_user_env(env_plugin: EnvironmentVariablePlugin, user: MockUser, results: tuple[tuple[str, str]]) -> None:
    with mock.patch.object(env_plugin, "_get_user_env_vars"):
        user_sid = None
        if user:
            user_sid = user.sid
        env_vars = env_plugin.user_env(user_sid)
        env_plugin._get_user_env_vars.assert_called_with(user_sid)
        assert env_vars == env_plugin._get_user_env_vars.return_value


def test_env(env_plugin: EnvironmentVariablePlugin) -> None:
    with mock.patch.object(env_plugin, "_get_system_env_vars"):
        env_vars = env_plugin.env
        env_plugin._get_system_env_vars.assert_called()
        assert env_vars == env_plugin._get_system_env_vars.return_value


def test_environment_variables(env_plugin: EnvironmentVariablePlugin) -> None:
    with (
        mock.patch.object(env_plugin, "_get_system_env_vars", side_effect=lambda: {"sys-name": "sys-value"}),
        mock.patch.object(env_plugin, "_get_user_env_vars", side_effect=lambda _: {"usr-name": "usr-value"}),
    ):
        records = list(env_plugin.environment_variables())
        # unwind the generator, so all functions are called
        list(records)
        env_plugin._get_system_env_vars.assert_called()
        env_plugin._get_user_env_vars.assert_called_once_with(MockUser.sid)
        assert records[0].name == "sys-name"
        assert records[0].value == "sys-value"
        assert records[0].username is None
        assert records[1].name == "usr-name"
        assert records[1].value == "usr-value"
        assert records[1].username == "test-user"


def test__get_pathext(env_plugin: EnvironmentVariablePlugin) -> None:
    pathexts = env_plugin.pathext
    pathexts = sorted(pathexts)
    assert pathexts == PATHEXTS


def test_pathext(env_plugin: EnvironmentVariablePlugin) -> None:
    with mock.patch.object(env_plugin, "_get_pathext"):
        env_plugin.pathext()
        env_plugin._get_pathext.assert_called_once()


def test_path_extensions(env_plugin: EnvironmentVariablePlugin) -> None:
    pathext_records = env_plugin.path_extensions()
    pathext_records = sorted(pathext_records, key=lambda record: record.pathext)
    for idx, record in enumerate(pathext_records):
        assert record.pathext == PATHEXTS[idx]



from __future__ import annotations

import os
import pathlib
import stat
import textwrap
from datetime import datetime, timezone
from io import BytesIO
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Any, BinaryIO
from unittest.mock import Mock, patch

import pytest
from dissect.util import ts

from dissect.target import filesystem
from dissect.target.exceptions import (
    FileNotFoundError,
    NotADirectoryError,
    SymlinkRecursionError,
)
from dissect.target.filesystem import (
    FilesystemEntry,
    LayerFilesystem,
    MappedFile,
    NotASymlinkError,
    RootFilesystem,
    RootFilesystemEntry,
    VirtualDirectory,
    VirtualFile,
    VirtualFilesystem,
    VirtualSymlink,
)
from dissect.target.filesystems.extfs import ExtFilesystem, ExtFilesystemEntry
from dissect.target.filesystems.ffs import FfsFilesystemEntry
from dissect.target.filesystems.tar import TarFilesystemEntry
from dissect.target.filesystems.xfs import XfsFilesystemEntry
from dissect.target.plugin import load_modules_from_paths

try:
    from dissect.target.filesystems.vmfs import VmfsFilesystemEntry
except ImportError:
    VmfsFilesystemEntry = None

from dissect.target.helpers import fsutil
from tests._utils import absolute_path


@pytest.fixture
def vfs() -> VirtualFilesystem:
    vfs = VirtualFilesystem()
    vfs.map_file_entry("/path/to/some/file", VirtualFile(vfs, "path/to/some/file", None))

    vfs.symlink("/path/to/some/", "dirlink1")
    vfs.symlink("dirlink1", "dirlink2")
    vfs.symlink("/path/to/some/file", "filelink1")
    vfs.symlink("filelink1", "filelink2")

    vfs.map_fs("/path/to/other", vfs)

    return vfs


def test_registration(tmp_path: Path) -> None:
    code = """
        from __future__ import annotations

        from typing import BinaryIO

        from dissect.target.filesystem import Filesystem, FilesystemEntry, register
        from dissect.target.volume import Volume


        class TestFilesystem(Filesystem):
            __type__ = "test"

            def __init__(self, case_sensitive: bool = True, alt_separator: str = None, volume: Volume = None):
                super().__init__(case_sensitive, alt_separator, volume)
                print("Helloworld from TestFilesystem")

            def get(self, path: str) -> FilesystemEntry:
                pass

            def detect(fh: BinaryIO) -> bool:
                return False


        register(__name__, TestFilesystem.__name__, internal=False)
    """

    (tmp_path / "filesystem.py").write_text(textwrap.dedent(code))

    with patch("dissect.target.filesystem.FILESYSTEMS", []) as mocked_filesystems:
        load_modules_from_paths([tmp_path])

        assert len(mocked_filesystems) == 1
        assert mocked_filesystems[0].__name__ == "TestFilesystem"


def test_get(vfs: VirtualFilesystem) -> None:
    assert vfs.get("dirlink1").name == "dirlink1"
    assert vfs.get("dirlink1").is_symlink()
    assert vfs.get("dirlink1").is_dir()
    assert not vfs.get("dirlink1").is_file()
    assert vfs.get("dirlink1").listdir() == ["file"]
    assert vfs.get("dirlink1").stat() != vfs.get("dirlink1").lstat()
    assert vfs.get("dirlink1").stat() == vfs.get("/path/to/some/").stat()

    assert vfs.get("dirlink2").name == "dirlink2"
    assert vfs.get("dirlink2").is_symlink()
    assert vfs.get("dirlink2").is_dir()
    assert not vfs.get("dirlink2").is_file()
    assert vfs.get("dirlink2").listdir() == ["file"]
    assert vfs.get("dirlink2").stat() != vfs.get("dirlink2").lstat()
    assert vfs.get("dirlink2").stat() == vfs.get("/path/to/some/").stat()

    assert vfs.get("filelink1").name == "filelink1"
    assert vfs.get("filelink1").is_symlink()
    assert not vfs.get("filelink1").is_dir()
    assert vfs.get("filelink1").is_file()
    assert vfs.get("filelink1").stat() != vfs.get("filelink1").lstat()
    assert vfs.get("filelink1").stat() == vfs.get("/path/to/some/file").stat()

    assert vfs.get("filelink2").name == "filelink2"
    assert vfs.get("filelink2").is_symlink()
    assert not vfs.get("filelink2").is_dir()
    assert vfs.get("filelink2").is_file()
    assert vfs.get("filelink2").stat() != vfs.get("filelink2").lstat()
    assert vfs.get("filelink2").stat() == vfs.get("/path/to/some/file").stat()


def test_symlink_across_layers() -> None:
    lfs = LayerFilesystem()

    vfs1 = VirtualFilesystem()
    vfs1.makedirs("/path/to/symlink/")
    vfs1.symlink("../target", "/path/to/symlink/target")

    vfs2 = VirtualFilesystem()
    target_dir = vfs2.makedirs("/path/to/target")

    layer1 = lfs.append_layer()
    layer1.mount("/", vfs1)

    layer2 = lfs.append_layer()
    layer2.mount("/", vfs2)

    target_entry = lfs.get("/path/to/symlink/target").readlink_ext()

    assert target_dir.stat() == target_entry.entries[0].stat()


def test_symlink_files_across_layers() -> None:
    lfs = LayerFilesystem()

    vfs1 = VirtualFilesystem()
    vfs1.makedirs("/path/to/symlink/")
    vfs1.symlink("../target", "/path/to/symlink/target")

    vfs2 = VirtualFilesystem()
    target_dir = vfs2.makedirs("/path/to/target/derp")

    layer1 = lfs.append_layer()
    layer1.mount("/", vfs1)

    layer2 = lfs.append_layer()
    layer2.mount("/", vfs2)

    target_entry = lfs.get("/path/to/symlink/target/derp")

    assert len(target_entry.entries) != 0
    assert target_dir.stat() == target_entry.stat()


def test_symlink_to_symlink_across_layers() -> None:
    lfs = LayerFilesystem()

    vfs1 = VirtualFilesystem()
    vfs1.makedirs("/path/to/symlink/")
    target_dir = vfs1.makedirs("/path/target")
    vfs1.symlink("../target", "/path/to/symlink/target")

    vfs2 = VirtualFilesystem()
    vfs2.symlink("../target", "/path/to/target")

    layer1 = lfs.append_layer()
    layer1.mount("/", vfs1)

    layer2 = lfs.append_layer()
    layer2.mount("/", vfs2)

    target_entry = lfs.get("/path/to/symlink/target/").readlink_ext()

    assert target_dir.stat() == target_entry.stat()


def test_recursive_symlink_across_layers() -> None:
    lfs = LayerFilesystem()

    vfs1 = VirtualFilesystem()
    vfs1.makedirs("/path/to/symlink/")
    vfs1.symlink("../target", "/path/to/symlink/target")

    vfs2 = VirtualFilesystem()
    vfs2.symlink("symlink/target", "/path/to/target")

    layer1 = lfs.append_layer()
    layer1.mount("/", vfs1)

    layer2 = lfs.append_layer()
    layer2.mount("/", vfs2)

    with pytest.raises(SymlinkRecursionError):
        lfs.get("/path/to/symlink/target/").readlink_ext()


def test_symlink_across_3_layers() -> None:
    lfs = LayerFilesystem()

    vfs1 = VirtualFilesystem()
    vfs1.makedirs("/path/to/symlink/")
    vfs1.symlink("../target", "/path/to/symlink/target")

    vfs2 = VirtualFilesystem()
    vfs2.symlink("../target", "/path/to/target")

    vfs3 = VirtualFilesystem()
    target_dir = vfs3.makedirs("/path/target")

    layer1 = lfs.append_layer()
    layer1.mount("/", vfs1)

    layer2 = lfs.append_layer()
    layer2.mount("/", vfs2)

    layer3 = lfs.append_layer()
    layer3.mount("/", vfs3)

    target_entry = lfs.get("/path/to/symlink/target/").readlink_ext()

    assert target_dir.stat() == target_entry.stat()
    stat_b = lfs.get("/path/to/symlink/target/").stat()
    stat_a = lfs.get("/path/to/target/").stat()
    assert stat_a == stat_b


def test_recursive_symlink_open_across_layers() -> None:
    lfs = LayerFilesystem()

    vfs1 = VirtualFilesystem()
    vfs1.makedirs("/path/to/symlink/")
    vfs1.symlink("../target", "/path/to/symlink/target")

    vfs2 = VirtualFilesystem()
    vfs2.symlink("symlink/target", "/path/to/target")

    layer1 = lfs.append_layer()
    layer1.mount("/", vfs1)

    layer2 = lfs.append_layer()
    layer2.mount("/", vfs2)

    with pytest.raises(SymlinkRecursionError):
        lfs.get("/path/to/symlink/target/").open()


def test_recursive_symlink_dev() -> None:
    lfs = LayerFilesystem()

    fs1 = ExtFilesystem(fh=absolute_path("_data/filesystems/symlink_disk.ext4").open("rb"))
    lfs.mount(fs=fs1, path="/")

    with pytest.raises(SymlinkRecursionError):
        lfs.get("/path/to/symlink/target/").readlink_ext()


@pytest.mark.parametrize(
    ("entry", "link_dict"),
    [
        (
            ExtFilesystemEntry,
            {"filetype": stat.S_IFLNK},
        ),
        (
            FfsFilesystemEntry,
            {},
        ),
        (
            VmfsFilesystemEntry,
            {},
        ),
        (
            XfsFilesystemEntry,
            {},
        ),
        (
            TarFilesystemEntry,
            {"linkname": "../target"},
        ),
    ],
)
def test_link_resolve(entry: type[FilesystemEntry], link_dict: dict[str, Any]) -> None:
    """Test wether each filesystem resolves a link as intended."""
    if entry is None:
        pytest.skip("dissect.vmfs is required")

    mocked_link = Mock()
    mocked_file = Mock()

    for k, v in link_dict.items():
        setattr(mocked_link, k, v)

    mocked_link.is_symlink.return_value = True
    mocked_link.link = "../target"

    mocked_file.is_symlink.return_value = False
    mocked_file.issym.return_value = False

    vfs = VirtualFilesystem()

    link = entry(vfs, "path/to/symlink/target", mocked_link)
    vfs.map_file_entry(vfspath="path/to/symlink/target", entry=link)

    actual_file = entry(vfs, "path/to/target", mocked_file)
    vfs.map_file_entry(vfspath="path/to/target", entry=actual_file)

    assert link.readlink_ext() == actual_file


def test_virtual_symlink_to_dir_get(vfs: VirtualFilesystem) -> None:
    some_file = vfs.get("/path/to/some/file")
    symlink = vfs.get("/dirlink1")

    some_file2 = symlink.get("file")

    assert some_file is some_file2


def test_virtual_symlink_to_file_get(vfs: VirtualFilesystem) -> None:
    symlink = vfs.get("/filelink1")
    with pytest.raises(NotADirectoryError):
        symlink.get("does_not_exist")


def test_virtual_symlink_to_symlink_get(vfs: VirtualFilesystem) -> None:
    some_file = vfs.get("/path/to/some/file")
    symlink = vfs.get("/dirlink2")

    some_file2 = symlink.get("file")

    assert some_file is some_file2


@pytest.mark.parametrize(
    "path",
    [
        "",
        "/",
    ],
)
@pytest.mark.parametrize(
    "entry_name",
    [
        "/path/to/some/file",
        "/path/to/some",
        "/dirlink1",
    ],
)
def test_virtual_entry_get_self(vfs: VirtualFilesystem, path: str, entry_name: str) -> None:
    some_entry = vfs.get(entry_name)
    some_entry2 = some_entry.get(path)

    assert some_entry is some_entry2


def test_virtual_filesystem_get() -> None:
    vfs = VirtualFilesystem()
    file_entry = VirtualFile(vfs, "file", None)
    vfs.map_file_entry("path/to/some/file", file_entry)

    assert vfs.get("") is vfs.root
    assert vfs.get("path/to/some/file") is file_entry


@pytest.mark.parametrize(
    ("vfs_path1", "vfs_path2"),
    [
        ("path/to/some/file", "/path/to/some/file"),
        ("path/to/some/file", "/path/to/some/file/"),
        ("path/to/some/file", "//path/to////some/file///"),
        ("path/to/some/file", "path/././to/some/./file"),
        ("path/to/some/file", "path/to/../to/some/../../../path/to/some/file"),
        ("path/to/some/file", "/dirlink1/file"),
        # as symlinks are resolved before continuing along the path, we have to
        # go back up 3 times to get back to the root
        ("path/to/some/file", "/dirlink1/../../../dirlink1/file"),
        ("path/to/some/file", "/dirlink1/../../../path/to/some/file"),
        ("path/to/some/file", "/dirlink2/file"),
        ("path/to/some/file", "/dirlink2/../../../dirlink1/file"),
        ("path/to/some/file", "/path/to/other/path/to/some/file"),
        ("path/to/some/file", "/path/to/other/dirlink1/file"),
        ("/", ""),
        ("/", "/path/to/../../"),
        ("/", "/dirlink2/../../../"),
    ],
)
def test_virtual_filesystem_get_equal_vfs_paths(vfs: VirtualFilesystem, vfs_path1: str, vfs_path2: str) -> None:
    assert vfs.get(vfs_path1) is vfs.get(vfs_path2)


@pytest.mark.parametrize(
    ("vfs_path1", "vfs_path2"),
    [
        ("path/to/some/file", "/filelink1"),
        ("path/to/some/file", "/filelink2"),
        ("/filelink1", "/filelink2"),
        ("path/to/some", "/dirlink1"),
        ("path/to/some", "/dirlink2"),
        ("/dirlink1", "/dirlink2"),
    ],
)
def test_virtual_filesystem_get_unequal_vfs_paths(vfs: VirtualFilesystem, vfs_path1: str, vfs_path2: str) -> None:
    assert vfs.get(vfs_path1) is not vfs.get(vfs_path2)


@pytest.mark.parametrize(
    ("vfs_path", "exception"),
    [
        ("/path/to/some/file/.", NotADirectoryError),
        ("/path/to/some/file/..", NotADirectoryError),
        ("/path/to/some/non-exisiting-file", FileNotFoundError),
        ("/path/to/other/path/to/some/non-exisiting-file", FileNotFoundError),
    ],
)
def test_virtual_filesystem_get_erroring_vfs_paths(
    vfs: VirtualFilesystem, vfs_path: str, exception: type[Exception]
) -> None:
    with pytest.raises(exception):
        vfs.get(vfs_path)


def test_virtual_filesystem_get_case_sensitive() -> None:
    vfs = VirtualFilesystem()
    vfs.map_file_entry("/path/to/some/file_lower_case", VirtualFile(vfs, "file_lower_case", None))
    vfs.map_file_entry("/path/TO/some/FILE_UPPER_CASE", VirtualFile(vfs, "FILE_UPPER_CASE", None))

    assert vfs.get("/path/to/some/file_lower_case").name == "file_lower_case"
    assert vfs.get("/path/TO/some/FILE_UPPER_CASE").name == "FILE_UPPER_CASE"
    with pytest.raises(FileNotFoundError):
        vfs.get("/path/to/some/FILE_LOWER_CASE")
    with pytest.raises(FileNotFoundError):
        assert vfs.get("/path/TO/some/file_upper_case")


def test_virtual_filesystem_get_case_insensitive() -> None:
    vfs = VirtualFilesystem(case_sensitive=False)
    vfs.map_file_entry("/path/to/some/file_lower_case", VirtualFile(vfs, "file_lower_case", None))
    vfs.map_file_entry("/path/TO/some/FILE_UPPER_CASE", VirtualFile(vfs, "FILE_UPPER_CASE", None))

    assert vfs.get("/path/to/some/file_lower_case").name == "file_lower_case"
    assert vfs.get("/path/TO/some/FILE_UPPER_CASE").name == "FILE_UPPER_CASE"
    assert vfs.get("/path/to/some/FILE_LOWER_CASE").name == "file_lower_case"
    assert vfs.get("/path/TO/some/file_upper_case").name == "FILE_UPPER_CASE"


@pytest.mark.parametrize(
    "paths",
    [
        ("/some/dir",),
        ("some/dir/",),
        ("/some/dir/",),
        ("some/dir",),
        (
            "/some/dir/",
            "some/dir/other/dir",
        ),
    ],
)
def test_virtual_filesystem_makedirs(paths: str) -> None:
    vfs = VirtualFilesystem()

    for vfspath in paths:
        vfs.makedirs(vfspath)

        partial_path = ""
        for part in vfspath.strip("/").split("/"):
            partial_path = fsutil.join(partial_path, part, alt_separator=vfs.alt_separator)
            vfs_entry = vfs.get(partial_path)

            assert isinstance(vfs_entry, VirtualDirectory)
            assert vfs_entry.path == partial_path.strip("/")


def test_virtual_filesystem_makedirs_root() -> None:
    vfs = VirtualFilesystem()
    vfspath = "/"

    vfs.makedirs(vfspath)

    vfs_entry = vfs.get(vfspath)

    assert vfs_entry is vfs.root


def test_virtual_filesystem_map_fs(vfs: VirtualFilesystem) -> None:
    root_vfs = VirtualFilesystem()
    map_path = "/some/dir/"
    file_path = "/path/to/some/file"
    root_file_path = fsutil.join(map_path, file_path, alt_separator=vfs.alt_separator)

    root_vfs.map_fs(map_path, vfs)

    vfs_entry = vfs.get(file_path)
    root_vfs_entry = vfs.get(root_file_path)

    assert vfs_entry is root_vfs_entry

    with pytest.raises(FileNotFoundError):
        root_vfs.get(file_path)


def test_virtual_filesystem_mount(vfs: VirtualFilesystem) -> None:
    assert vfs.mount == vfs.map_fs


def test_virtual_filesystem_map_dir(tmp_path: Path) -> None:
    vfs = VirtualFilesystem()
    vfs_path = "/map/point/"
    with (
        TemporaryDirectory(dir=tmp_path) as tmp_dir,
        TemporaryDirectory(dir=tmp_dir) as some_dir,
        TemporaryDirectory(dir=tmp_dir) as other_dir,
        TemporaryDirectory(dir=other_dir) as second_lvl_dir,
        NamedTemporaryFile(dir=some_dir, delete=False) as some_file,
    ):
        some_file.write(b"1337")
        some_file.close()

        vfs.map_dir(vfs_path, tmp_dir)

        rel_path = os.path.relpath(some_dir, tmp_dir)
        rel_path = fsutil.normalize(rel_path, alt_separator=os.path.sep)
        entry_name = fsutil.join(vfs_path, rel_path, alt_separator=vfs.alt_separator)
        dir_entry = vfs.get(entry_name)
        assert isinstance(dir_entry, VirtualDirectory)

        rel_path = os.path.relpath(second_lvl_dir, tmp_dir)
        rel_path = fsutil.normalize(rel_path, alt_separator=os.path.sep)
        entry_name = fsutil.join(vfs_path, rel_path, alt_separator=vfs.alt_separator)
        dir_entry = vfs.get(entry_name)
        assert isinstance(dir_entry, VirtualDirectory)

        rel_path = os.path.relpath(some_file.name, tmp_dir)
        rel_path = fsutil.normalize(rel_path, alt_separator=os.path.sep)
        entry_name = fsutil.join(vfs_path, rel_path, alt_separator=vfs.alt_separator)
        file_entry = vfs.get(entry_name)
        assert isinstance(file_entry, MappedFile)

        fp = file_entry.open()
        assert fp.read() == b"1337"
        fp.close()


@pytest.mark.parametrize(
    "vfs_path",
    [
        "/path/to/file",
        "path/to/file",
        "/path///to/file",
    ],
)
def test_virtual_filesystem_map_file(vfs_path: str) -> None:
    vfs = VirtualFilesystem()
    real_path = "/tmp/foo"

    vfs.map_file(vfs_path, real_path)

    vfs_path = fsutil.normalize(vfs_path, alt_separator=vfs.alt_separator).strip("/")
    vfs_entry = vfs.get(vfs_path)

    assert isinstance(vfs_entry, MappedFile)
    assert vfs_entry.path == vfs_path
    assert vfs_entry.entry == real_path


def test_virtual_filesystem_map_file_as_dir() -> None:
    vfs = VirtualFilesystem()
    real_path = "/tmp/foo"

    with pytest.raises(AttributeError):
        vfs.map_file("/path/to/dir/", real_path)


@pytest.mark.parametrize(
    "vfs_path",
    [
        "/path/to/file",
        "path/to/file",
        "/path///to/file",
    ],
)
def test_virtual_filesystem_map_file_fh(vfs_path: str) -> None:
    vfs = VirtualFilesystem()
    fh = Mock()

    vfs.map_file_fh(vfs_path, fh)

    vfs_path = fsutil.normalize(vfs_path, alt_separator=vfs.alt_separator).strip("/")
    vfs_entry = vfs.get(vfs_path)

    assert isinstance(vfs_entry, VirtualFile)
    assert vfs_entry.path == vfs_path
    assert vfs_entry.entry is fh


def test_virtual_filesystem_map_file_fh_as_dir() -> None:
    vfs = VirtualFilesystem()
    fh = Mock()

    with pytest.raises(AttributeError):
        vfs.map_file_fh("/path/to/dir/", fh)


@pytest.mark.parametrize(
    "vfs_path",
    [
        "/path/to/entry",
        "path/to/entry",
        "path/to/entry/",
        "/path/to/entry/",
        "//path///to/entry//",
        "/entry",
        "entry",
        "entry/",
        "/entry/",
        "/",
    ],
)
def test_virtual_filesystem_map_file_entry(vfs_path: str) -> None:
    vfs = VirtualFilesystem()
    entry_path = fsutil.normalize(vfs_path, alt_separator=vfs.alt_separator).strip("/")
    dir_entry = VirtualDirectory(vfs, entry_path)

    file_name = fsutil.join(vfs_path, "test", alt_separator=vfs.alt_separator)
    file_entry = VirtualFile(vfs, file_name, Mock())
    dir_entry.add("test", file_entry)

    vfs.map_file_entry(vfs_path, dir_entry)

    vfs_entry = vfs.get(vfs_path)

    if vfs_path == "/":
        assert vfs.get(file_name) is file_entry
    else:
        assert vfs_entry is dir_entry


@pytest.mark.parametrize(
    ("vfs_path", "link_path"),
    [
        (
            "/path/to/entry",
            "/path/to/link",
        ),
        (
            "path/to/entry",
            "path/to/link",
        ),
        (
            "path/to/entry/",
            "path/to/link/",
        ),
        (
            "/path/to/entry/",
            "/path/to/link/",
        ),
        (
            "//path///to/entry//",
            "//path///to/link//",
        ),
    ],
)
def test_virtual_filesystem_link(vfs_path: str, link_path: str) -> None:
    vfs = VirtualFilesystem()
    entry_path = fsutil.normalize(vfs_path, alt_separator=vfs.alt_separator).strip("/")
    file_object = Mock()
    file_entry = VirtualFile(vfs, entry_path, file_object)
    vfs.map_file_entry(vfs_path, file_entry)

    vfs.link(vfs_path, link_path)

    link_path = fsutil.normalize(link_path, alt_separator=vfs.alt_separator).strip("/")
    link_entry = vfs.get(link_path)

    assert link_entry is file_entry


@pytest.mark.parametrize(
    ("vfs_path", "link_path"),
    [
        (
            "/path/to/entry",
            "/path/to/link",
        ),
        (
            "path/to/entry",
            "path/to/link",
        ),
        (
            "path/to/entry/",
            "path/to/link/",
        ),
        (
            "/path/to/entry/",
            "/path/to/link/",
        ),
        (
            "//path///to/entry//",
            "//path///to/link//",
        ),
    ],
)
def test_virtual_filesystem_symlink(vfs_path: str, link_path: str) -> None:
    vfs = VirtualFilesystem()

    vfs.symlink(vfs_path, link_path)

    vfs_path = fsutil.normalize(vfs_path, alt_separator=vfs.alt_separator).rstrip("/")
    link_path = fsutil.normalize(link_path, alt_separator=vfs.alt_separator).strip("/")
    link_entry = vfs.get(link_path)

    assert isinstance(link_entry, VirtualSymlink)
    assert link_entry.path == link_path
    assert link_entry.target == vfs_path


@pytest.fixture
def dir_entry(vfs: VirtualFilesystem) -> VirtualDirectory:
    return vfs.get("/path/to/some")


@pytest.fixture
def dirlink_entry(vfs: VirtualFilesystem) -> VirtualSymlink:
    return vfs.get("dirlink1")


@pytest.fixture
def file_entry(vfs: VirtualFilesystem) -> VirtualFile:
    return vfs.get("/path/to/some/file")


@pytest.fixture
def filelink_entry(vfs: VirtualFilesystem) -> VirtualSymlink:
    return vfs.get("filelink1")


@pytest.mark.parametrize(
    ("src_entry", "dst_entry"),
    [
        ("dir_entry", "dir_entry"),
        ("dirlink_entry", "dir_entry"),
        ("file_entry", "file_entry"),
        ("filelink_entry", "file_entry"),
    ],
)
def test_virtual_filesystem_stat(
    vfs: VirtualFilesystem,
    src_entry: str,
    dst_entry: str,
    request: pytest.FixtureRequest,
) -> None:
    src_entry = request.getfixturevalue(src_entry)
    dst_entry = request.getfixturevalue(dst_entry)

    assert vfs.stat(src_entry.path, follow_symlinks=False) == src_entry.stat(follow_symlinks=False)
    assert vfs.stat(src_entry.path, follow_symlinks=True) == dst_entry.stat(follow_symlinks=True)


@pytest.mark.parametrize(
    "entry",
    [
        "dir_entry",
        "dirlink_entry",
        "file_entry",
        "filelink_entry",
    ],
)
def test_virtual_filesystem_lstat(vfs: VirtualFilesystem, entry: str, request: pytest.FixtureRequest) -> None:
    entry = request.getfixturevalue(entry)

    assert vfs.lstat(entry.path) == entry.lstat()
    assert vfs.stat(entry.path, follow_symlinks=False) == entry.lstat()


@pytest.mark.parametrize(
    ("src_entry", "dst_entry"),
    [
        ("dir_entry", "dir_entry"),
        ("dirlink_entry", "dir_entry"),
        ("file_entry", "file_entry"),
        ("filelink_entry", "file_entry"),
    ],
)
def test_virtual_filesystem_is_dir(
    vfs: VirtualFilesystem,
    src_entry: str,
    dst_entry: str,
    request: pytest.FixtureRequest,
) -> None:
    src_entry = request.getfixturevalue(src_entry)
    dst_entry = request.getfixturevalue(dst_entry)

    assert vfs.is_dir(src_entry.path, follow_symlinks=False) == src_entry.is_dir(follow_symlinks=False)
    assert vfs.is_dir(src_entry.path, follow_symlinks=True) == src_entry.is_dir(follow_symlinks=True)
    assert vfs.is_dir(src_entry.path, follow_symlinks=True) == dst_entry.is_dir(follow_symlinks=False)
    assert vfs.is_dir(src_entry.path, follow_symlinks=True) == dst_entry.is_dir(follow_symlinks=True)


@pytest.mark.parametrize(
    ("src_entry", "dst_entry"),
    [
        ("dir_entry", "dir_entry"),
        ("dirlink_entry", "dir_entry"),
        ("file_entry", "file_entry"),
        ("filelink_entry", "file_entry"),
    ],
)
def test_virtual_filesystem_is_file(
    vfs: VirtualFilesystem,
    src_entry: str,
    dst_entry: str,
    request: pytest.FixtureRequest,
) -> None:
    src_entry = request.getfixturevalue(src_entry)
    dst_entry = request.getfixturevalue(dst_entry)

    assert vfs.is_file(src_entry.path, follow_symlinks=False) == src_entry.is_file(follow_symlinks=False)
    assert vfs.is_file(src_entry.path, follow_symlinks=True) == src_entry.is_file(follow_symlinks=True)
    assert vfs.is_file(src_entry.path, follow_symlinks=True) == dst_entry.is_file(follow_symlinks=False)
    assert vfs.is_file(src_entry.path, follow_symlinks=True) == dst_entry.is_file(follow_symlinks=True)


@pytest.fixture
def virt_dir() -> VirtualDirectory:
    return VirtualDirectory(Mock(), "")


@pytest.fixture
def top_virt_dir() -> VirtualDirectory:
    return VirtualDirectory(Mock(), "")


def test_virtual_directory_stat(virt_dir: VirtualDirectory, top_virt_dir: VirtualDirectory) -> None:
    assert virt_dir.stat(follow_symlinks=False) == virt_dir._stat()
    assert virt_dir.stat(follow_symlinks=True) == virt_dir._stat()

    virt_dir.top = top_virt_dir
    assert virt_dir.stat(follow_symlinks=False) == top_virt_dir.stat(follow_symlinks=False)
    assert virt_dir.stat(follow_symlinks=True) == top_virt_dir.stat(follow_symlinks=True)


def test_virtual_directory_lstat(virt_dir: VirtualDirectory, top_virt_dir: VirtualDirectory) -> None:
    assert virt_dir.lstat() == virt_dir._stat()
    assert virt_dir.lstat() == virt_dir.stat(follow_symlinks=False)
    assert virt_dir.lstat().st_mode == stat.S_IFDIR

    virt_dir.top = top_virt_dir
    assert virt_dir.lstat() == top_virt_dir.lstat()


def test_virtual_directory_is_dir(virt_dir: VirtualDirectory) -> None:
    assert virt_dir.is_dir(follow_symlinks=True)
    assert virt_dir.is_dir(follow_symlinks=False)


def test_virtual_directory_is_file(virt_dir: VirtualDirectory) -> None:
    assert not virt_dir.is_file(follow_symlinks=True)
    assert not virt_dir.is_file(follow_symlinks=False)


@pytest.fixture
def virt_file() -> VirtualFile:
    return VirtualFile(Mock(), "", Mock())


def test_virtual_file_stat(virt_file: VirtualFile) -> None:
    assert virt_file.stat(follow_symlinks=False) == virt_file.lstat()
    assert virt_file.stat(follow_symlinks=True) == virt_file.lstat()


def test_virtual_file_lstat(virt_file: VirtualFile) -> None:
    assert virt_file.lstat().st_mode == stat.S_IFREG


def test_virtual_file_is_dir(virt_file: VirtualFile) -> None:
    assert not virt_file.is_dir(follow_symlinks=True)
    assert not virt_file.is_dir(follow_symlinks=False)


def test_virtual_file_is_file(virt_file: VirtualFile) -> None:
    assert virt_file.is_file(follow_symlinks=True)
    assert virt_file.is_file(follow_symlinks=False)


def test_virtual_symlink_stat(filelink_entry: VirtualSymlink, file_entry: VirtualFile | VirtualDirectory) -> None:
    assert filelink_entry.stat(follow_symlinks=False) == filelink_entry.lstat()
    assert filelink_entry.stat(follow_symlinks=True) == file_entry.stat()


def test_virtual_symlink_lstat(filelink_entry: VirtualSymlink) -> None:
    assert filelink_entry.lstat().st_mode == stat.S_IFLNK


@pytest.mark.parametrize(
    ("virt_link", "is_dir"),
    [
        ("dirlink_entry", True),
        ("filelink_entry", False),
    ],
)
def test_virtual_symlink_is_dir(virt_link: str, is_dir: bool, request: pytest.FixtureRequest) -> None:
    virt_link = request.getfixturevalue(virt_link)

    assert virt_link.is_dir(follow_symlinks=False) is False
    assert virt_link.is_dir(follow_symlinks=True) == is_dir


@pytest.mark.parametrize(
    ("virt_link", "is_file"),
    [
        ("dirlink_entry", False),
        ("filelink_entry", True),
    ],
)
def test_virtual_symlink_is_file(virt_link: str, is_file: bool, request: pytest.FixtureRequest) -> None:
    virt_link = request.getfixturevalue(virt_link)

    assert virt_link.is_file(follow_symlinks=False) is False
    assert virt_link.is_file(follow_symlinks=True) == is_file


@pytest.fixture
def vfs1() -> VirtualFilesystem:
    return VirtualFilesystem()


@pytest.fixture
def vfs2() -> VirtualFilesystem:
    return VirtualFilesystem()


@pytest.fixture
def vfs1_entry(vfs1: VirtualFilesystem) -> VirtualFile:
    return VirtualFile(vfs1, "vfs1_entry", Mock())


@pytest.fixture
def vfs2_entry(vfs2: VirtualFilesystem) -> VirtualFile:
    return VirtualFile(vfs2, "vfs2_entry", Mock())


@pytest.fixture
def rootfs(
    vfs1: VirtualFilesystem,
    vfs2: VirtualFilesystem,
    vfs1_entry: VirtualFile,
    vfs2_entry: VirtualFile,
) -> RootFilesystem:
    vfs1.map_file_entry("/vfs1_entry", vfs1_entry)
    vfs1.symlink("/vfs1/vfs2/", "/link_to_vfs2")

    vfs2.map_file_entry("/vfs2_entry", vfs2_entry)
    vfs2_shared_entry = VirtualFile(vfs2, "shared_entry", Mock())
    vfs2.map_file_entry("/shared_entry", vfs2_shared_entry)

    vfs1.map_file_entry("/path/to/some/file", VirtualFile(vfs1, "path/to/some/file", Mock()))
    vfs1.symlink("path/to/some/", "dirlink")
    vfs1.symlink("path/to/some/file", "filelink")

    target = Mock()
    rootfs = RootFilesystem(target)
    rootfs.mount("/vfs1", vfs1)
    rootfs.mount("/vfs1/vfs2", vfs2)
    return rootfs


def test_root_filesystem_get(
    rootfs: RootFilesystem,
    vfs1: VirtualFilesystem,
    vfs1_entry: VirtualFile,
    vfs2_entry: VirtualFile,
) -> None:
    entry_path = "/vfs1/vfs1_entry"
    rootfs_entry = rootfs.get(entry_path)
    assert rootfs_entry.path == entry_path
    assert len(rootfs_entry.entries) == 1
    assert vfs1_entry in rootfs_entry.entries

    entry_path = "/vfs1/vfs2/vfs2_entry"
    rootfs_entry = rootfs.get(entry_path)
    assert rootfs_entry.path == entry_path
    assert len(rootfs_entry.entries) == 1
    assert vfs2_entry in rootfs_entry.entries

    entry_path = "/vfs1"
    rootfs_entry = rootfs.get(entry_path)
    assert rootfs_entry.path == entry_path
    assert len(rootfs_entry.entries) == 2
    assert isinstance(rootfs_entry.entries[0], VirtualDirectory)
    assert isinstance(rootfs_entry.entries[1], VirtualDirectory)

    entry_path = "/vfs1/link_to_vfs2"
    rootfs_entry = rootfs.get(entry_path)
    vfs1_link = vfs1.get("/link_to_vfs2")
    assert rootfs_entry.path == entry_path
    assert len(rootfs_entry.entries) == 1
    assert vfs1_link in rootfs_entry.entries

    entry_path = "/vfs1/link_to_vfs2/vfs2_entry"
    rootfs_entry = rootfs.get(entry_path)
    assert rootfs_entry.path == entry_path
    assert len(rootfs_entry.entries) == 1
    nested_rootfs_entry = rootfs_entry.entries[0]
    assert isinstance(nested_rootfs_entry, RootFilesystemEntry)
    assert len(nested_rootfs_entry.entries) == 1
    assert vfs2_entry in nested_rootfs_entry.entries


@pytest.fixture
def root_dir_entry(rootfs: RootFilesystem) -> VirtualDirectory:
    return rootfs.get("/vfs1/path/to/some")


@pytest.fixture
def root_dirlink_entry(rootfs: RootFilesystem) -> VirtualSymlink:
    return rootfs.get("/vfs1/dirlink")


@pytest.fixture
def root_file_entry(rootfs: RootFilesystem) -> VirtualFile:
    return rootfs.get("/vfs1/path/to/some/file")


@pytest.fixture
def root_filelink_entry(rootfs: RootFilesystem) -> VirtualSymlink:
    return rootfs.get("/vfs1/filelink")


@pytest.mark.parametrize(
    ("src_entry", "dst_entry"),
    [
        ("root_dir_entry", "root_dir_entry"),
        ("root_dirlink_entry", "root_dir_entry"),
        ("root_file_entry", "root_file_entry"),
        ("root_filelink_entry", "root_file_entry"),
    ],
)
def test_root_filesystem_entry_stat(src_entry: str, dst_entry: str, request: pytest.FixtureRequest) -> None:
    src_entry = request.getfixturevalue(src_entry)
    dst_entry = request.getfixturevalue(dst_entry)

    assert src_entry.stat(follow_symlinks=False) == src_entry.lstat()
    assert src_entry.stat(follow_symlinks=True) == dst_entry.stat()


@pytest.mark.parametrize(
    ("entry", "st_mode"),
    [
        ("root_dir_entry", stat.S_IFDIR),
        ("root_dirlink_entry", stat.S_IFLNK),
        ("root_file_entry", stat.S_IFREG),
    ],
)
def test_root_filesystem_entry_lstat(entry: str, st_mode: int, request: pytest.FixtureRequest) -> None:
    entry = request.getfixturevalue(entry)

    assert entry.lstat().st_mode == st_mode


@pytest.mark.parametrize(
    ("entry", "src_is_dir", "dst_is_dir"),
    [
        ("root_dir_entry", True, True),
        ("root_dirlink_entry", False, True),
        ("root_file_entry", False, False),
        ("root_filelink_entry", False, False),
    ],
)
def test_root_filesystem_entry_is_dir(
    entry: str, src_is_dir: bool, dst_is_dir: bool, request: pytest.FixtureRequest
) -> None:
    entry = request.getfixturevalue(entry)

    assert entry.is_dir(follow_symlinks=False) == src_is_dir
    assert entry.is_dir(follow_symlinks=True) == dst_is_dir


@pytest.mark.parametrize(
    ("entry", "src_is_file", "dst_is_file"),
    [
        ("root_dir_entry", False, False),
        ("root_dirlink_entry", False, False),
        ("root_file_entry", True, True),
        ("root_filelink_entry", False, True),
    ],
)
def test_root_filesystem_entry_is_file(
    entry: str,
    src_is_file: bool,
    dst_is_file: bool,
    request: pytest.FixtureRequest,
) -> None:
    entry = request.getfixturevalue(entry)

    assert entry.is_file(follow_symlinks=False) == src_is_file
    assert entry.is_file(follow_symlinks=True) == dst_is_file


@pytest.fixture
def mapped_file() -> MappedFile:
    return MappedFile(Mock(), "/some/path", "/host/path")


def test_mapped_file_stat(mapped_file: MappedFile) -> None:
    mock_stat = Mock()

    with (
        patch("dissect.target.helpers.fsutil.stat_result.copy", autospec=True) as stat_copy,
        patch("pathlib.Path.stat", autospec=True, return_value=mock_stat) as path_stat,
        patch("pathlib.Path.lstat", autospec=True, return_value=mock_stat) as path_lstat,
    ):
        mapped_file.stat(follow_symlinks=False)
        path_lstat.assert_called_with(pathlib.Path(mapped_file.entry))
        stat_copy.assert_called_with(mock_stat)

        mapped_file.stat(follow_symlinks=True)
        path_stat.assert_called_with(pathlib.Path(mapped_file.entry))
        stat_copy.assert_called_with(mock_stat)


def test_mapped_file_lstat(mapped_file: MappedFile) -> None:
    mock_stat = Mock()

    with (
        patch("dissect.target.helpers.fsutil.stat_result.copy", autospec=True) as stat_copy,
        patch("pathlib.Path.lstat", autospec=True, return_value=mock_stat) as os_lstat,
    ):
        mapped_file.lstat()
        os_lstat.assert_called_with(pathlib.Path(mapped_file.entry))
        stat_copy.assert_called_with(mock_stat)


def test_mapped_file_attr(mapped_file: MappedFile) -> None:
    with patch("dissect.target.helpers.fsutil.fs_attrs", autospec=True) as fs_attrs:
        mapped_file.attr()
        fs_attrs.assert_called_with(mapped_file.entry, follow_symlinks=True)


def test_mapped_file_lattr(mapped_file: MappedFile) -> None:
    with patch("dissect.target.helpers.fsutil.fs_attrs", autospec=True) as fs_attrs:
        mapped_file.lattr()
        fs_attrs.assert_called_with(mapped_file.entry, follow_symlinks=False)


def test_mapped_file_is_symlink(vfs: VirtualFilesystem) -> None:
    assert MappedFile(vfs, "/guestlink", "/hostlink").is_symlink() is False


def test_mapped_file_readlink(vfs: VirtualFilesystem) -> None:
    with pytest.raises(NotASymlinkError):
        MappedFile(vfs, "/guestlink", "/hostlink").readlink()
    with pytest.raises(NotASymlinkError):
        MappedFile(vfs, "/guestlink", "/hostlink").readlink_ext()


def test_reset_file_position() -> None:
    fh = BytesIO(b"\x00" * 8192)
    fh.seek(512)

    class MockFilesystem(filesystem.Filesystem):
        def __init__(self, fh: BinaryIO):
            assert fh.tell() == 0
            fh.seek(1024)
            self.success = True

        @staticmethod
        def _detect(fh: BinaryIO) -> bool:
            assert fh.tell() == 0
            fh.seek(256)
            return True

    mock_fs = Mock()
    mock_fs.MockFilesystem = MockFilesystem

    with patch.object(filesystem, "FILESYSTEMS", [mock_fs.MockFilesystem]):
        opened_fs = filesystem.open(fh)
        assert isinstance(opened_fs, mock_fs.MockFilesystem)
        assert opened_fs.success
        assert fh.tell() == 512


def test_virtual_filesystem_map_dir_from_tar() -> None:
    mock_fs = VirtualFilesystem()
    tar_file = absolute_path("_data/loaders/tar/test-archive.tar")
    mock_fs.map_dir_from_tar("/foo/bar", tar_file)

    assert mock_fs.path("/foo/bar/test-data/test-file.txt").exists()
    assert mock_fs.path("/foo/bar/test-data/test-file.txt").open().read() == b"test-value\n"

    stat = mock_fs.path("/foo/bar/test-data/test-file.txt").stat()
    assert ts.from_unix(stat.st_mtime) == datetime(2021, 12, 6, 9, 51, 40, tzinfo=timezone.utc)


def test_virtual_filesystem_map_file_from_tar() -> None:
    mock_fs = VirtualFilesystem()
    tar_file = absolute_path("_data/loaders/tar/test-archive.tar")
    mock_fs.map_file_from_tar("/var/example/test.txt", tar_file)

    assert mock_fs.path("/var/example/test.txt").exists()
    assert mock_fs.path("/var/example/test.txt").open().read() == b"test-value\n"

    stat = mock_fs.path("/var/example/test.txt").stat()
    assert ts.from_unix(stat.st_mtime) == datetime(2021, 12, 6, 9, 51, 40, tzinfo=timezone.utc)


def test_layer_filesystem() -> None:
    lfs = LayerFilesystem()

    vfs1 = VirtualFilesystem()
    vfs1.map_file_fh("file1", BytesIO(b"value1"))

    vfs2 = VirtualFilesystem()
    vfs2.map_file_fh("file2", BytesIO(b"value2"))

    vfs3 = VirtualFilesystem()
    vfs3.map_file_fh("file3", BytesIO(b"value3"))

    vfs4 = VirtualFilesystem()
    vfs4.map_file_fh("file1", BytesIO(b"value4"))

    lfs.append_fs_layer(vfs1)
    assert lfs.path("file1").read_text() == "value1"

    lfs.append_fs_layer(vfs2)
    assert lfs.path("file1").read_text() == "value1"
    assert lfs.path("file2").read_text() == "value2"

    lfs.append_fs_layer(vfs3)
    assert lfs.path("file1").read_text() == "value1"
    assert lfs.path("file2").read_text() == "value2"
    assert lfs.path("file3").read_text() == "value3"

    lfs.append_fs_layer(vfs4)
    assert lfs.path("file1").read_text() == "value4"
    lfs.remove_fs_layer(vfs4)
    lfs.prepend_fs_layer(vfs4)
    assert lfs.path("file1").read_text() == "value1"


def test_layer_filesystem_mount() -> None:
    lfs = LayerFilesystem()

    vfs1 = VirtualFilesystem()
    vfs1.map_file_fh("file1", BytesIO(b"value1"))

    vfs2 = VirtualFilesystem()
    vfs2.map_file_fh("file2", BytesIO(b"value2"))

    lfs.mount("/vfs", vfs1)
    lfs.mount("/vfs", vfs2, ignore_existing=True)

    assert lfs.listdir("/vfs") == ["file2"]

    lfs.mount("/vfs", vfs1, ignore_existing=True)

    assert lfs.listdir("/vfs") == ["file1"]

    lfs.mount("/vfs", vfs2, ignore_existing=False)

    assert sorted(lfs.listdir("/vfs")) == ["file1", "file2"]
    assert lfs.path("/vfs/file1").read_text() == "value1"
    assert lfs.path("/vfs/file2").read_text() == "value2"


def test_layer_filesystem_relative_link() -> None:
    """Test relative symlinks from a filesystem mounted at a subdirectory."""
    lfs = LayerFilesystem()

    vfs = VirtualFilesystem()
    vfs.map_file_fh("dir/hello/world", BytesIO(b"o/"))
    vfs.symlink("dir/hello", "bye")

    lfs.mount("/mnt", vfs)

    assert lfs.path("/mnt/bye/world").read_text() == "o/"



from __future__ import annotations

import hashlib
import logging
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from dissect.cstruct import cstruct
from dissect.util.sid import read_sid

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers import keychain
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.windows.dpapi.crypto import (
    CipherAlgorithm,
    HashAlgorithm,
    derive_password_hash,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

log = logging.getLogger(__name__)


CredHistRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/credential/history",
    [
        ("string", "guid"),
        ("boolean", "decrypted"),
        ("string", "sha1"),
        ("string", "nt"),
    ],
)


credhist_def = """
struct entry {
    DWORD   dwVersion;
    CHAR    guidLink[16];
    DWORD   dwNextLinkSize;
    DWORD   dwCredLinkType;
    DWORD   algHash;                    // ALG_ID
    DWORD   dwPbkdf2IterationCount;
    DWORD   dwSidSize;
    DWORD   algCrypt;                   // ALG_ID
    DWORD   dwShaHashSize;
    DWORD   dwNtHashSize;
    CHAR    pSalt[16];
    CHAR    pSid[dwSidSize];
    CHAR    encrypted[0];
};
"""

c_credhist = cstruct().load(credhist_def)


@dataclass
class CredHistEntry:
    version: int
    guid: str
    user_sid: str
    sha1: bytes | None
    nt: bytes | None
    hash_alg: HashAlgorithm = field(repr=False)
    cipher_alg: CipherAlgorithm = field(repr=False)
    raw: c_credhist.entry = field(repr=False)
    decrypted: bool = False

    def decrypt(self, password_hash: bytes) -> None:
        """Decrypt this CREDHIST entry using the provided password hash. Modifies ``CredHistEntry.sha1``
        and ``CredHistEntry.nt`` values.

        If the decrypted ``nt`` value is 16 bytes we assume the decryption was successful.

        Args:
            password_hash: Bytes of SHA1 password hash digest.

        Raises:
            ValueError: If the decryption seems to have failed.
        """
        data = self.cipher_alg.decrypt_with_hmac(
            data=self.raw.encrypted,
            key=derive_password_hash(password_hash, self.user_sid),
            iv=self.raw.pSalt,
            hash_algorithm=self.hash_alg,
            rounds=self.raw.dwPbkdf2IterationCount,
        )

        sha_size = self.raw.dwShaHashSize
        nt_size = self.raw.dwNtHashSize

        sha1 = data[:sha_size]
        nt = data[sha_size : sha_size + nt_size].rstrip(b"\x00")

        if len(nt) != 16:
            raise ValueError("Decrypting failed, invalid password hash?")

        self.decrypted = True
        self.sha1 = sha1
        self.nt = nt


class CredHistFile:
    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.entries = list(self._parse())

    def __repr__(self) -> str:
        return f"<CredHistFile fh='{self.fh}' entries={len(self.entries)}>"

    def _parse(self) -> Iterator[CredHistEntry]:
        self.fh.seek(0)
        try:
            while True:
                entry = c_credhist.entry(self.fh)

                # determine size of encrypted data and add to entry
                cipher_alg = CipherAlgorithm.from_id(entry.algCrypt)
                enc_size = entry.dwShaHashSize + entry.dwNtHashSize
                enc_size += enc_size % cipher_alg.block_length
                entry.encrypted = self.fh.read(enc_size)

                yield CredHistEntry(
                    version=entry.dwVersion,
                    guid=UUID(bytes_le=entry.guidLink),
                    user_sid=read_sid(entry.pSid) if entry.pSid else None,
                    hash_alg=HashAlgorithm.from_id(entry.algHash),
                    cipher_alg=cipher_alg,
                    sha1=None,
                    nt=None,
                    raw=entry,
                )
        except EOFError:
            # An empty CREDHIST file will be 24 bytes long and has dwNextLinkSize set to 0.
            pass

    def decrypt(self, password_hash: bytes) -> None:
        """Decrypt a CREDHIST chain using the provided password SHA1 hash."""

        for entry in reversed(self.entries):
            try:
                entry.decrypt(password_hash)
            except ValueError as e:
                log.warning("Could not decrypt entry %s with password %s", entry.guid, password_hash.hex())
                log.debug("", exc_info=e)
                continue
            password_hash = entry.sha1


class CredHistPlugin(Plugin):
    """Windows CREDHIST file parser.

    Windows XP:         ``C:\\Documents and Settings\\username\\Application Data\\Microsoft\\Protect\\CREDHIST``
    Windows 7 and up:   ``C:\\Users\\username\\AppData\\Roaming\\Microsoft\\Protect\\CREDHIST``

    References:
        - https://www.passcape.com/index.php?section=docsys&cmd=details&id=28#41
    """

    def __init__(self, target: Target):
        super().__init__(target)
        self.files = list(self._find_files())

    def _find_files(self) -> Iterator[tuple[UserDetails, Path]]:
        hashes = set()
        for user_details in self.target.user_details.all_with_home():
            for path in ["AppData/Roaming/Microsoft/Protect/CREDHIST", "Application Data/Microsoft/Protect/CREDHIST"]:
                credhist_path = user_details.home_path.joinpath(path)
                if credhist_path.exists() and (hash := credhist_path.get().hash()) not in hashes:
                    hashes.add(hash)
                    yield user_details.user, credhist_path

    def check_compatible(self) -> None:
        if not self.files:
            raise UnsupportedPluginError("No CREDHIST files found on target.")

    @export(record=CredHistRecord)
    def credhist(self) -> Iterator[CredHistRecord]:
        """Yield and decrypt all Windows CREDHIST entries on the target."""

        passwords = keychain_passwords()

        if not passwords:
            self.target.log.warning("No passwords provided in keychain, cannot decrypt CREDHIST hashes")

        for user, path in self.files:
            credhist = CredHistFile(path.open("rb"))

            for password in passwords:
                credhist.decrypt(hashlib.sha1(password.encode("utf-16-le")).digest())

            for entry in credhist.entries:
                yield CredHistRecord(
                    guid=entry.guid,
                    decrypted=entry.decrypted,
                    sha1=entry.sha1.hex() if entry.sha1 else None,
                    nt=entry.nt.hex() if entry.nt else None,
                    _user=user,
                    _target=self.target,
                )


def keychain_passwords() -> set:
    passphrases = set()
    for key in keychain.get_keys_for_provider("user") + keychain.get_keys_without_provider():
        if key.key_type == keychain.KeyType.PASSPHRASE:
            passphrases.add(key.value)
    passphrases.add("")
    return passphrases



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.volume import disk

from dissect.target.volume import Volume, VolumeSystem

if TYPE_CHECKING:
    from collections.abc import Iterator


class DissectVolumeSystem(VolumeSystem):
    __type__ = "disk"

    def __init__(self, fh: BinaryIO | list[BinaryIO], *args, **kwargs):
        self._disk = disk.Disk(fh)
        super().__init__(fh, *args, serial=self._disk.serial, **kwargs)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        try:
            disk.Disk(fh)
        except Exception:
            return False
        else:
            return True

    def _volumes(self) -> Iterator[Volume]:
        for v in self._disk.partitions:
            name = v.name or f"part_{v.offset:08x}"
            yield Volume(
                v.open(),
                v.number,
                v.offset,
                v.size,
                v.type,
                name,
                guid=v.guid,
                raw=v,
                disk=self.disk,
                vs=self,
            )



from __future__ import annotations

import gzip
import io
import logging
import zlib
from itertools import cycle, islice
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import AlignedStream, RangeStream, RelativeStream

from dissect.target.container import Container
from dissect.target.tools.utils import catch_sigpipe

if TYPE_CHECKING:
    from pathlib import Path

log = logging.getLogger(__name__)


def find_xor_key(fh: BinaryIO) -> bytes:
    """Find the XOR key for the firmware file by using known plaintext of zeros.

    File-like object ``fh`` must be seeked to the correct offset where it should decode to all zeroes (0x00).

    Arguments:
        fh: File-like object to read from.

    Returns:
        bytes: XOR key, note that the XOR key is not validated and may be incorrect.
    """
    key = bytearray()

    pos = fh.tell()
    buf = fh.read(32)
    fh.seek(pos)

    if pos % 512 == 0:
        xor_char = 0xFF
    else:
        fh.seek(pos - 1)
        xor_char = ord(fh.read(1))

    for i, k_char in enumerate(buf):
        idx = (i + pos) & 0x1F
        key.append((xor_char ^ k_char ^ idx) & 0xFF)
        xor_char = k_char

    # align xor key
    koffset = 32 - (pos & 0x1F)
    key = islice(cycle(key), koffset, koffset + 32)
    return bytes(key)


class FortiFirmwareFile(AlignedStream):
    """Fortinet firmware file, handles transparant decompression and deobfuscation of the firmware file."""

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.trailer_offset = None
        self.trailer_data = None
        self.xor_key = None
        self.is_gzipped = False

        size = None

        # Check if the file is gzipped
        self.fh.seek(0)
        header = self.fh.read(4)
        if header.startswith(b"\x1f\x8b"):
            self.is_gzipped = True

            # Find the extra metadata behind the gzip compressed data
            # as a bonus we can also calculate the size of the firmware here
            dec = zlib.decompressobj(wbits=16 + zlib.MAX_WBITS)
            self.fh.seek(0)
            size = 0
            while True:
                data = self.fh.read(io.DEFAULT_BUFFER_SIZE)
                if not data:
                    break
                d = dec.decompress(dec.unconsumed_tail + data)
                size += len(d)

            # Ignore the trailer data of the gzip file if we have any
            if dec.unused_data:
                self.trailer_offset = self.fh.seek(-len(dec.unused_data), io.SEEK_END)
                self.trailer_data = self.fh.read()
                log.debug("Found trailer offset: %d, data: %r", self.trailer_offset, self.trailer_data)
                self.fh = RangeStream(self.fh, 0, self.trailer_offset)

            self.fh.seek(0)
            self.fh = gzip.GzipFile(fileobj=self.fh)

        # Find the xor key based on known offsets where the firmware should decode to zero bytes
        for zero_offset in (0x30, 0x40, 0x400):
            self.fh.seek(zero_offset)
            xor_key = find_xor_key(self.fh)
            if xor_key.isalnum():
                self.xor_key = xor_key
                log.info("Found key %r @ offset %s", self.xor_key, zero_offset)
                break
        else:
            log.info("No xor key found")

        # Determine the size of the firmware file if we didn't calculate it yet
        if size is None:
            size = self.fh.seek(0, io.SEEK_END)

        log.info("firmware size: %s", size)
        log.info("xor key: %r", self.xor_key)
        log.info("gzipped: %s", self.is_gzipped)
        self.fh.seek(0)

        # Align the stream to 512 bytes which simplifies the XOR deobfuscation code
        super().__init__(size=size, align=512)

    def _read(self, offset: int, length: int) -> bytes:
        self.fh.seek(offset)
        buf = self.fh.read(length)

        if not self.xor_key:
            return buf

        buf = bytearray(buf)
        xor_char = 0xFF
        for i, cur_char in enumerate(buf):
            if (i + offset) % 512 == 0:
                xor_char = 0xFF
            idx = (i + offset) & 0x1F
            buf[i] = ((self.xor_key[idx] ^ cur_char ^ xor_char) - idx) & 0xFF
            xor_char = cur_char

        return bytes(buf)


class FortiFirmwareContainer(Container):
    __type__ = "fortifw"

    def __init__(self, fh: BinaryIO | Path, *args, **kwargs):
        if not hasattr(fh, "read"):
            fh = fh.open("rb")

        # Open the firmware file
        self.ff = FortiFirmwareFile(fh)

        # seek to MBR
        self.fw = RelativeStream(self.ff, 0x200)
        super().__init__(self.fw, self.ff.size, *args, **kwargs)

    @staticmethod
    def detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return False

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        # all Fortinet firmware files end with `-FORTINET.out`
        return str(path).lower().endswith("-fortinet.out")

    def read(self, length: int) -> bytes:
        return self.fw.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.fw.seek(offset, whence)

    def tell(self) -> int:
        return self.fw.tell()

    def close(self) -> None:
        pass


@catch_sigpipe
def main(argv: list[str] | None = None) -> None:
    import argparse
    import shutil
    import sys

    parser = argparse.ArgumentParser(description="Decompress and deobfuscate Fortinet firmware file to stdout.")
    parser.add_argument("file", type=argparse.FileType("rb"), help="Fortinet firmware file")
    parser.add_argument("--verbose", "-v", action="store_true", help="verbose output")
    args = parser.parse_args(argv)

    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    ff = FortiFirmwareFile(args.file)
    shutil.copyfileobj(ff, sys.stdout.buffer)


if __name__ == "__main__":
    main()



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows.wer import WindowsErrorReportingPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_wer_plugin(target_win: Target, fs_win: VirtualFilesystem) -> None:
    wer_dir = absolute_path("_data/plugins/os/windows/wer")
    fs_win.map_dir("ProgramData/Microsoft/Windows/WER/ReportQueue/test", wer_dir)
    target_win.add_plugin(WindowsErrorReportingPlugin)
    tests = [
        "os_version_information_lcid",
        "response_type",
        "sig",
        "dynamic_sig",
        "dynamic_signatures_parameter1",
        "ui_x5b_1_x5d",
        "sp_xc3a9_cial_charact_xc3a9_r",
        "xd0bd_xd0b5_xd0b2_xd0b8_xd0b4_xd0b8_xd0bc_xd18b_xd0b9",
        "x5f_start_with_an_",
        "x33__start_with_a_3",
    ]

    records = list(target_win.wer())
    assert len(records) == 2

    wer_record_map = {r.wer_file_path.name: r for r in records}
    assert "wer_test.wer" in wer_record_map
    assert "wer_test_no_bom.wer" in wer_record_map

    record = wer_record_map["wer_test.wer"]
    for test in tests:
        record_field = getattr(record, test, None)
        assert record_field == f"test_{test}"

    assert record.ts == datetime.datetime(2022, 10, 4, 11, 0, 0, 0, tzinfo=datetime.timezone.utc)

    record = wer_record_map["wer_test_no_bom.wer"]
    assert record.ts == datetime.datetime(2021, 11, 22, 15, 39, 49, 20733, tzinfo=datetime.timezone.utc)
    assert record.os_version == "6.1.7601.2.1.0.256.48"
    assert record.app_name == "Microsoft Malware Protection Command Line Utility"



from __future__ import annotations

from typing import TYPE_CHECKING, Union, get_args

from dissect.target.exceptions import RegistryError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, alias, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import RegistryKey

PanelPathRecord = TargetRecordDescriptor(
    "application/productivity/sevenzip/panelpath",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

ArcHistoryRecord = TargetRecordDescriptor(
    "application/productivity/sevenzip/archistory",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

PathHistoryRecord = TargetRecordDescriptor(
    "application/productivity/sevenzip/pathhistory",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

CopyHistoryRecord = TargetRecordDescriptor(
    "application/productivity/sevenzip/copyhistory",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

FolderHistoryRecord = TargetRecordDescriptor(
    "application/productivity/sevenzip/folderhistory",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

SevenZipRecord = Union[PanelPathRecord, ArcHistoryRecord, PathHistoryRecord, CopyHistoryRecord, FolderHistoryRecord]


class SevenZipPlugin(Plugin):
    """Windows 7-Zip GUI plugin."""

    KEY = "HKCU\\Software\\7-Zip"

    def check_compatible(self) -> None:
        if not self.target.has_function("registry") or not list(self.target.registry.keys(self.KEY)):
            raise UnsupportedPluginError("7-Zip registry key not found")

    def parse_key(
        self, key: RegistryKey, keyname: str, valuename: str, record: TargetRecordDescriptor
    ) -> Iterator[TargetRecordDescriptor]:
        try:
            subkey = key.subkey(keyname)
            value = subkey.value(valuename).value
            for file_path in value.decode("utf-16-le").split("\x00"):
                if not file_path:
                    continue

                yield record(
                    ts=subkey.ts,
                    path=self.target.fs.path(file_path),
                    _target=self.target,
                )
        except RegistryError:
            pass

    @export(record=get_args(SevenZipRecord))
    @alias("7zip")
    def sevenzip(self) -> Iterator[SevenZipRecord]:
        """Return 7-Zip GUI history information from the registry.

        7-Zip is an open source file archiver. If the HKCU\\Software\\7-Zip registry key exists, it checks for
        additional registry keys, such as ArcHistory and FolderHistory. This might provide insight in which files have
        been archived by the 7-Zip GUI.

        References:
            - https://www.7-zip.org/
        """
        target = self.target
        for key in target.registry.keys(self.KEY):
            try:
                subkey = key.subkey("FM")
                value = subkey.value("PanelPath0").value
                yield PanelPathRecord(
                    ts=subkey.ts,
                    path=self.target.fs.path(value),
                    _target=self.target,
                )
            except RegistryError:
                pass

            yield from self.parse_key(key, "Compression", "ArcHistory", ArcHistoryRecord)
            yield from self.parse_key(key, "Extraction", "PathHistory", PathHistoryRecord)
            yield from self.parse_key(key, "FM", "CopyHistory", CopyHistoryRecord)
            yield from self.parse_key(key, "FM", "FolderHistory", FolderHistoryRecord)



from __future__ import annotations

import plistlib
from functools import cache, lru_cache
from typing import TYPE_CHECKING

from dissect.target.exceptions import FileNotFoundError
from dissect.target.helpers.record import MacOSInterfaceRecord
from dissect.target.plugins.os.default.network import NetworkPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target


class MacOSNetworkPlugin(NetworkPlugin):
    """macOS network interface plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self._plistnetwork = cache(self._plistnetwork)
        self._plistlease = lru_cache(32)(self._plistlease)

    def _plistlease(self, devname: str) -> dict:
        for lease in self.target.fs.glob_ext(f"/private/var/db/dhcpclient/leases/{devname}*"):
            return plistlib.load(lease.open())
        return {}

    def _plistnetwork(self) -> dict:
        if (preferences := self.target.fs.path("/Library/Preferences/SystemConfiguration/preferences.plist")).exists():
            return plistlib.load(preferences.open()) or {}

        raise FileNotFoundError("Couldn't find preferences file")

    def _interfaces(self) -> Iterator[MacOSInterfaceRecord]:
        plistnetwork = self._plistnetwork()
        current_set = plistnetwork.get("CurrentSet")
        sets = plistnetwork.get("Sets", {})
        for name, _set in sets.items():
            if f"/Sets/{name}" == current_set:
                item = _set
                for key in ["Network", "Global", "IPv4", "ServiceOrder"]:
                    item = item.get(key, {})
                service_order = item
                break

        network = plistnetwork.get("NetworkServices", {})
        vlans = plistnetwork.get("VirtualNetworkInterfaces", {}).get("VLAN", {})

        vlan_lookup = {key: vlan.get("Tag") for key, vlan in vlans.items()}

        for _id, interface in network.items():
            dns = set()
            gateways = set()
            device = interface.get("Interface", {})
            name = device.get("DeviceName")
            _type = device.get("Type")
            vlan = vlan_lookup.get(name)
            dhcp = False
            ifaces = set()
            interface_service_order = service_order.index(_id) if _id in service_order else None
            try:
                for addr in interface.get("DNS", {}).get("ServerAddresses", {}):
                    dns.add(addr)

                for addresses in [interface.get("IPv4", {}), interface.get("IPv6", {})]:
                    iface = ""
                    if router := addresses.get("Router"):
                        gateways.add(router)

                    if addresses.get("ConfigMethod", "") == "DHCP":
                        iface = self._plistlease(name).get("IPAddress")
                        dhcp = True
                    elif addr := addresses.get("Addresses", []):
                        iface = addr[0]

                    if _subnet_mask := list(filter(None, addresses.get("SubnetMasks", []))):
                        iface = f"{iface}/{_subnet_mask[0]}"

                    if iface:
                        ifaces.add(iface)

                yield MacOSInterfaceRecord(
                    name=name,
                    type=_type,
                    enabled=not interface.get("__INACTIVE__", False),
                    cidr=ifaces,
                    gateway=list(gateways),
                    dns=list(dns),
                    mac=[],
                    source="NetworkServices",
                    interface_service_order=interface_service_order,
                    dhcp=dhcp,
                    vlan=vlan,
                    _target=self.target,
                )

            except Exception as e:
                self.target.log.warning("Error reading configuration for network device %s", name)
                self.target.log.debug("", exc_info=e)



from __future__ import annotations

import datetime
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any, TextIO

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export
from dissect.target.plugins.os.windows.defender.mplog import (
    DEFENDER_MPLOG_BLOCK_PATTERNS,
    DEFENDER_MPLOG_LINE,
    DEFENDER_MPLOG_PATTERNS,
    DefenderMPLogBMTelemetryRecord,
    DefenderMPLogDetectionAddRecord,
    DefenderMPLogDetectionEventRecord,
    DefenderMPLogEMSRecord,
    DefenderMPLogExclusionRecord,
    DefenderMPLogLowfiRecord,
    DefenderMPLogMinFilBlockedFileRecord,
    DefenderMPLogMinFilUSSRecord,
    DefenderMPLogOriginalFileNameRecord,
    DefenderMPLogProcessImageRecord,
    DefenderMPLogResourceScanRecord,
    DefenderMPLogRTPRecord,
    DefenderMPLogThreatActionRecord,
    DefenderMPLogThreatRecord,
)
from dissect.target.plugins.os.windows.defender.quarantine import (
    DefenderFileQuarantineRecord,
    DefenderQuarantineRecord,
    QuarantineEntry,
    recover_quarantined_file_streams,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from flow.record import Record

DEFENDER_EVTX_FIELDS = [
    ("datetime", "ts"),
    ("uint32", "EventID"),
    ("string", "Provider_Name"),
    ("string", "Action_ID"),
    ("string", "Action_Name"),
    ("string", "Additional_Actions_ID"),
    ("string", "Additional_Actions_String"),
    ("string", "Category_ID"),
    ("string", "Category_Name"),
    ("string", "Channel"),
    ("string", "Computer"),
    ("string", "Correlation_ActivityID"),
    ("string", "Correlation_RelatedActivityID"),
    ("string", "Detection_ID"),
    ("datetime", "Detection_Time"),
    ("string", "Detection_User"),
    ("string", "Engine_Version"),
    ("string", "Error_Code"),
    ("string", "Error_Description"),
    ("string", "EventID_Qualifiers"),
    ("string", "EventRecordID"),
    ("string", "Execution_ID"),
    ("string", "Execution_Name"),
    ("string", "Execution_ProcessID"),
    ("string", "Execution_ThreadID"),
    ("string", "FWLink"),
    ("string", "Keywords"),
    ("string", "Level"),
    ("string", "Opcode"),
    ("string", "Origin_ID"),
    ("string", "Origin_Name"),
    ("string", "Path"),
    ("string", "Post_Clean_Status"),
    ("string", "Pre_Execution_Status"),
    ("string", "Process_Name"),
    ("string", "Product_Name"),
    ("string", "Product_Version"),
    ("string", "Provider_Guid"),
    ("string", "Remediation_User"),
    ("string", "Security_intelligence_Version"),
    ("string", "Security_UserID"),
    ("string", "Severity_ID"),
    ("string", "Severity_Name"),
    ("string", "Source_ID"),
    ("string", "Source_Name"),
    ("string", "State"),
    ("string", "Status_Code"),
    ("string", "Status_Description"),
    ("string", "Task"),
    ("string", "Threat_ID"),
    ("string", "Threat_Name"),
    ("string", "Type_ID"),
    ("string", "Type_Name"),
    ("string", "Version"),
]

DEFENDER_LOG_DIR = "%windir%/system32/winevt/logs"
DEFENDER_LOG_FILENAME_GLOB = "Microsoft-Windows-Windows Defender*"
EVTX_PROVIDER_NAME = "Microsoft-Windows-Windows Defender"

DEFENDER_QUARANTINE_DIR = "sysvol/programdata/microsoft/windows defender/quarantine"
DEFENDER_MPLOG_DIR = "sysvol/programdata/microsoft/windows defender/support"
DEFENDER_KNOWN_DETECTION_TYPES = [b"internalbehavior", b"regkey", b"runkey"]

DEFENDER_EXCLUSION_KEY = "HKLM\\SOFTWARE\\Microsoft\\Windows Defender\\Exclusions"

DefenderLogRecord = TargetRecordDescriptor(
    "filesystem/windows/defender/evtx",
    DEFENDER_EVTX_FIELDS,
)

DefenderExclusionRecord = TargetRecordDescriptor(
    "filesystem/windows/defender/exclusion",
    [
        ("datetime", "regf_mtime"),
        ("string", "type"),
        ("string", "value"),
    ],
)


def parse_iso_datetime(datetime_value: str) -> datetime.datetime:
    """Parse ISO8601 serialized datetime with ``Z`` ending."""
    return datetime.datetime.strptime(datetime_value, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=datetime.timezone.utc)


def filter_records(records: Iterable, field_name: str, field_value: Any) -> Iterator[DefenderLogRecord]:
    """Apply a filter on an Iterable of records, returning only records that have the given field value for the given
    field name.
    """

    def filter_func(record: Record) -> bool:
        return hasattr(record, field_name) and getattr(record, field_name) == field_value

    return filter(filter_func, records)


class MicrosoftDefenderPlugin(Plugin):
    """Plugin that parses artifacts created by Microsoft Defender.

    This includes the EVTX logs, as well as recovery of artefacts from the quarantine folder.
    """

    __namespace__ = "defender"

    def check_compatible(self) -> None:
        # Either the Defender log folder, the quarantine folder or the exclusions registry key
        # has to exist for this plugin to be compatible.

        if not any(
            [
                self.target.resolve(DEFENDER_LOG_DIR).exists(),
                self.target.fs.path(DEFENDER_QUARANTINE_DIR).exists(),
                (
                    self.target.has_function("registry")
                    and len(list(self.target.registry.keys(DEFENDER_EXCLUSION_KEY))) > 0
                ),
            ]
        ):
            raise UnsupportedPluginError("No Defender objects found")

    @export(record=DefenderLogRecord)
    def evtx(self) -> Iterator[DefenderLogRecord]:
        """Parse Microsoft Defender evtx log files."""

        defender_evtx_field_names = [field_name for _, field_name in DEFENDER_EVTX_FIELDS]

        evtx_records = self.target.evtx(logs_dir=DEFENDER_LOG_DIR, log_file_glob=DEFENDER_LOG_FILENAME_GLOB)
        defender_evtx_records = filter_records(evtx_records, "Provider_Name", "Microsoft-Windows-Windows Defender")

        for evtx_record in defender_evtx_records:
            record_fields = {}
            for field_name in defender_evtx_field_names:
                if not hasattr(evtx_record, field_name):
                    continue

                value = getattr(evtx_record, field_name)

                if field_name == "Detection_Time" and value:
                    value = parse_iso_datetime(value)

                record_fields[field_name] = value

            yield DefenderLogRecord(**record_fields, _target=self.target)

    @export(record=[DefenderQuarantineRecord, DefenderFileQuarantineRecord])
    def quarantine(self) -> Iterator[DefenderQuarantineRecord | DefenderFileQuarantineRecord]:
        """Parse the quarantine folder of Microsoft Defender for quarantine entry resources.

        Quarantine entry resources contain metadata about detected threats that Microsoft Defender has placed in
        quarantine.
        """
        for entry in self.get_quarantine_entries():
            # These fields are present for all (currently known) quarantine entry types
            fields = {
                "ts": entry.timestamp,
                "quarantine_id": entry.quarantine_id.hex(),
                "scan_id": entry.scan_id.hex(),
                "threat_id": entry.threat_id,
                "detection_name": entry.detection_name,
            }
            for resource in entry.resources:
                fields.update({"detection_type": resource.detection_type})
                if resource.detection_type == b"file":
                    # These fields are only available for file based detections
                    fields.update(
                        {
                            "detection_path": resource.detection_path,
                            "creation_time": resource.creation_time,
                            "last_write_time": resource.last_write_time,
                            "last_accessed_time": resource.last_access_time,
                            "resource_id": resource.resource_id,
                        }
                    )
                    yield DefenderFileQuarantineRecord(**fields, _target=self.target)
                else:
                    # For these types, we know that they have no known additional data to add to the Quarantine Record.
                    if resource.detection_type not in DEFENDER_KNOWN_DETECTION_TYPES:
                        self.target.log.warning(
                            "Unknown Defender Detection Type %s, yielding a generic quarantine record.",
                            resource.detection_type,
                        )
                    # For anything other than a file, we yield a generic DefenderQuarantineRecord.
                    yield DefenderQuarantineRecord(**fields, _target=self.target)

    @export(record=DefenderExclusionRecord)
    def exclusions(self) -> Iterator[DefenderExclusionRecord]:
        """Yield Microsoft Defender exclusions from the Registry."""

        # Iterate through all possible versions of the key for Defender exclusions
        for exclusions_registry_key in self.target.registry.keys(DEFENDER_EXCLUSION_KEY):
            # Every subkey of the exclusions key is a 'type' of exclusion, e.g. 'path' 'process' or 'extension'.
            for exclusion_type_subkey in exclusions_registry_key.subkeys():
                # Every value is an exclusion for a said type. The 'name' property of said registry value holds
                # what is being excluded from Defender (e.g. powershell.exe, notepad.txt)
                for exclusion in exclusion_type_subkey.values():
                    exclusion_value = exclusion.name
                    exclusion_type = exclusion_type_subkey.name
                    # Due to the fact that every exclusion is a registry value and not a registry key, we can only know
                    # the last modified timestamp of the exclusion type for a given exclusion, not a timestamp for the
                    # exclusion itself. We reflect this to the analyst by using the regf_mtime field.
                    yield DefenderExclusionRecord(
                        regf_mtime=exclusion_type_subkey.timestamp,
                        type=exclusion_type,
                        value=exclusion_value,
                        _target=self.target,
                    )

    def _mplog_processimage(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogProcessImageRecord]:
        yield DefenderMPLogProcessImageRecord(**data)

    def _mplog_minfiluss(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogMinFilUSSRecord]:
        yield DefenderMPLogMinFilUSSRecord(**data)

    def _mplog_blockedfile(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogMinFilBlockedFileRecord]:
        yield DefenderMPLogMinFilBlockedFileRecord(**data)

    def _mplog_bmtelemetry(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogBMTelemetryRecord]:
        data["ts"] = datetime.datetime.strptime(data["ts"], "%m-%d-%Y %H:%M:%S").replace(tzinfo=tzinfo)
        yield DefenderMPLogBMTelemetryRecord(**data)

    def _mplog_ems(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogEMSRecord]:
        yield DefenderMPLogEMSRecord(**data)

    def _mplog_originalfilename(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogOriginalFileNameRecord]:
        yield DefenderMPLogOriginalFileNameRecord(**data)

    def _mplog_exclusion(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogExclusionRecord]:
        yield DefenderMPLogExclusionRecord(**data)

    def _mplog_lowfi(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogLowfiRecord]:
        yield DefenderMPLogLowfiRecord(**data)

    def _mplog_detectionadd(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogDetectionAddRecord]:
        yield DefenderMPLogDetectionAddRecord(**data)

    def _mplog_threat(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogThreatRecord]:
        yield DefenderMPLogThreatRecord(**data)

    def _mplog_resourcescan(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogResourceScanRecord]:
        data["start_time"] = datetime.datetime.strptime(data["start_time"], "%m-%d-%Y %H:%M:%S").replace(tzinfo=tzinfo)
        data["end_time"] = datetime.datetime.strptime(data["end_time"], "%m-%d-%Y %H:%M:%S").replace(tzinfo=tzinfo)
        data["ts"] = data["start_time"]
        rest = data.pop("rest")
        yield DefenderMPLogResourceScanRecord(
            threats=re.findall("Threat Name:([^\n]+)", rest),
            resources=re.findall("Resource Path:([^\n]+)", rest),
            **data,
        )

    def _mplog_threataction(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogThreatActionRecord]:
        data["ts"] = datetime.datetime.strptime(data["ts"], "%m-%d-%Y %H:%M:%S").replace(tzinfo=tzinfo)
        rest = data.pop("rest")
        yield DefenderMPLogThreatActionRecord(
            threats=re.findall("Threat Name:([^\n]+)", rest),
            resources=re.findall("(?:Path|File Name):([^\n]+)", rest),
            actions=re.findall("Action:([^\n]+)", rest),
            **data,
        )

    def _mplog_rtp_log(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogRTPRecord]:
        times = {}
        for dtkey in ["ts", "last_perf", "first_rtp_scan"]:
            try:
                times[dtkey] = datetime.datetime.strptime(data[dtkey], "%m-%d-%Y %H:%M:%S").replace(tzinfo=tzinfo)
            except ValueError:  # noqa: PERF203
                pass

        yield DefenderMPLogRTPRecord(
            _target=self.target,
            source_log=data["source_log"],
            **times,
            plugin_states=re.findall(r"^\s+(.*)$", data["plugin_states"])[0],
            process_exclusions=re.findall(DEFENDER_MPLOG_LINE, data["process_exclusions"]),
            path_exclusions=re.findall(DEFENDER_MPLOG_LINE, data["path_exclusions"]),
            ext_exclusions=re.findall(DEFENDER_MPLOG_LINE, data["ext_exclusions"]),
        )

    def _mplog_detectionevent(
        self, data: dict, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogDetectionEventRecord]:
        yield DefenderMPLogDetectionEventRecord(**data)

    def _mplog_line(
        self, mplog_line: str, source: Path, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[
        DefenderMPLogProcessImageRecord
        | DefenderMPLogMinFilUSSRecord
        | DefenderMPLogMinFilBlockedFileRecord
        | DefenderMPLogEMSRecord
        | DefenderMPLogOriginalFileNameRecord
        | DefenderMPLogExclusionRecord
        | DefenderMPLogLowfiRecord
        | DefenderMPLogDetectionAddRecord
        | DefenderMPLogThreatRecord
        | DefenderMPLogDetectionEventRecord
    ]:
        for pattern, record in DEFENDER_MPLOG_PATTERNS:
            if match := pattern.match(mplog_line):
                data = match.groupdict()
                data["_target"] = self.target
                data["source_log"] = source
                yield from getattr(self, f"_mplog_{record.name.split('/')[-1:][0]}")(data, tzinfo=tzinfo)

    def _mplog_block(
        self, mplog_line: str, mplog: TextIO, source: Path, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[DefenderMPLogResourceScanRecord | DefenderMPLogThreatActionRecord | DefenderMPLogRTPRecord]:
        block = ""
        for prefix, suffix, pattern, record in DEFENDER_MPLOG_BLOCK_PATTERNS:
            if prefix.search(mplog_line):
                block += mplog_line

                while mplog_line := mplog.readline():
                    block += mplog_line
                    if suffix.search(mplog_line):
                        break
                match = pattern.match(block)
                if not match:
                    return

                data = match.groupdict()
                data["_target"] = self.target
                data["source_log"] = source
                yield from getattr(self, f"_mplog_{record.name.split('/')[-1:][0]}")(data, tzinfo=tzinfo)

    def _mplog(
        self, mplog: TextIO, source: Path, tzinfo: datetime.tzinfo = datetime.timezone.utc
    ) -> Iterator[
        DefenderMPLogProcessImageRecord
        | DefenderMPLogMinFilUSSRecord
        | DefenderMPLogMinFilBlockedFileRecord
        | DefenderMPLogBMTelemetryRecord
        | DefenderMPLogEMSRecord
        | DefenderMPLogOriginalFileNameRecord
        | DefenderMPLogExclusionRecord
        | DefenderMPLogLowfiRecord
        | DefenderMPLogDetectionAddRecord
        | DefenderMPLogThreatRecord
        | DefenderMPLogDetectionEventRecord
        | DefenderMPLogResourceScanRecord
        | DefenderMPLogThreatActionRecord
        | DefenderMPLogRTPRecord
    ]:
        while mplog_line := mplog.readline():
            yield from self._mplog_line(mplog_line, source, tzinfo=tzinfo)
            yield from self._mplog_block(mplog_line, mplog, source, tzinfo=tzinfo)

    @export(
        record=[
            DefenderMPLogProcessImageRecord,
            DefenderMPLogMinFilUSSRecord,
            DefenderMPLogMinFilBlockedFileRecord,
            DefenderMPLogBMTelemetryRecord,
            DefenderMPLogEMSRecord,
            DefenderMPLogOriginalFileNameRecord,
            DefenderMPLogExclusionRecord,
            DefenderMPLogLowfiRecord,
            DefenderMPLogDetectionAddRecord,
            DefenderMPLogThreatRecord,
            DefenderMPLogDetectionEventRecord,
            DefenderMPLogResourceScanRecord,
            DefenderMPLogThreatActionRecord,
            DefenderMPLogRTPRecord,
        ]
    )
    def mplog(
        self,
    ) -> Iterator[
        DefenderMPLogProcessImageRecord
        | DefenderMPLogMinFilUSSRecord
        | DefenderMPLogMinFilBlockedFileRecord
        | DefenderMPLogBMTelemetryRecord
        | DefenderMPLogEMSRecord
        | DefenderMPLogOriginalFileNameRecord
        | DefenderMPLogExclusionRecord
        | DefenderMPLogLowfiRecord
        | DefenderMPLogDetectionAddRecord
        | DefenderMPLogThreatRecord
        | DefenderMPLogDetectionEventRecord
        | DefenderMPLogResourceScanRecord
        | DefenderMPLogThreatActionRecord
        | DefenderMPLogRTPRecord
    ]:
        """Return the contents of the Defender MPLog file.

        References:
            - https://www.crowdstrike.com/blog/how-to-use-microsoft-protection-logging-for-forensic-investigations/
            - https://www.intrinsec.com/hunt-mplogs/
            - https://github.com/Intrinsec/mplog_parser
        """
        target_tz = self.target.datetime.tzinfo

        mplog_directory = self.target.fs.path(DEFENDER_MPLOG_DIR)

        if not (mplog_directory.exists() and mplog_directory.is_dir()):
            return

        for mplog_file in mplog_directory.glob("MPLog-*"):
            for encoding in ["UTF-16", "UTF-8"]:
                try:
                    with mplog_file.open("rt", encoding=encoding) as mplog:
                        yield from self._mplog(mplog, self.target.fs.path(mplog_file), tzinfo=target_tz)
                    break
                except UnicodeError:
                    continue

    @arg(
        "-o",
        "--output",
        dest="output_dir",
        type=Path,
        required=True,
        help="Path to recover quarantined file to.",
    )
    @export(output="none")
    def recover(self, output_dir: Path) -> None:
        """Recover files that have been placed into quarantine by Microsoft Defender.

        Microsoft Defender RC4 encrypts the output of the 'BackupRead' function when it places a file into quarantine.
        This means multiple data streams can be contained in a single quarantined file, including zone identifier
        information.
        """
        if not output_dir.exists():
            raise ValueError("Output directory does not exist")

        quarantine_directory = self.target.fs.path(DEFENDER_QUARANTINE_DIR)
        resourcedata_directory = quarantine_directory.joinpath("ResourceData")
        if resourcedata_directory.exists() and resourcedata_directory.is_dir():
            recovered_files = []
            for entry in self.get_quarantine_entries():
                for resource in entry.resources:
                    if resource.detection_type != b"file":
                        # We can only recover file entries
                        continue
                    # First two characters of the resource ID is the subdirectory that will contain the quarantined file
                    subdir = resource.resource_id[0:2]
                    resourcedata_location = resourcedata_directory.joinpath(subdir).joinpath(resource.resource_id)
                    if not resourcedata_location.exists():
                        self.target.log.warning("Could not find a ResourceData file for %s", resource.resource_id)
                        continue
                    if not resourcedata_location.is_file():
                        self.target.log.warning("%s is not a file!", resourcedata_location)
                        continue
                    if resourcedata_location in recovered_files:
                        # We already recovered this file
                        continue
                    with resourcedata_location.open() as fh:
                        # We restore the file with the resource_id as its filename. While we could 'guess' the filename
                        # based on the information we have from the associated quarantine entry, there is a potential
                        # that different files have the same filename. Analysts can use the quarantine records to cross
                        # reference.
                        for dest_filename, dest_buf in recover_quarantined_file_streams(fh, resource.resource_id):
                            output_filename = output_dir.joinpath(dest_filename)
                            self.target.log.info("Saving %s", output_filename)
                            output_filename.write_bytes(dest_buf)

                    # Make sure we do not recover the same file multiple times if it has multiple entries
                    recovered_files.append(resourcedata_location)

    def get_quarantine_entries(self) -> Iterator[QuarantineEntry]:
        """Yield Windows Defender quarantine entries."""
        quarantine_directory = self.target.fs.path(DEFENDER_QUARANTINE_DIR)
        entries_directory = quarantine_directory.joinpath("entries")

        if not entries_directory.is_dir():
            return
        for guid_path in entries_directory.iterdir():
            if not guid_path.exists() or not guid_path.is_file():
                continue
            with guid_path.open() as entry_fh:
                entry = QuarantineEntry(entry_fh)

            # Warn on discovery of fields that we do not have knowledge of what they are / do.
            for resource in entry.resources:
                for unknown_field in resource.unknown_fields:
                    self.target.log.warning("Encountered an unknown field identifier: %s", unknown_field.Identifier)
            yield entry



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

ConsoleHostHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "powershell/history",
    [
        ("datetime", "mtime"),
        ("varint", "order"),
        ("string", "command"),
        ("path", "source"),
    ],
)


class PowerShellHistoryPlugin(Plugin):
    """Windows PowerShell history plugin."""

    PATHS = (
        "AppData/Roaming/Microsoft/Windows/PowerShell/psreadline",
        ".local/share/powershell/PSReadLine",
    )

    def __init__(self, target: Target):
        super().__init__(target)

        self._history = []

        for user_details in target.user_details.all_with_home():
            for ps_path in self.PATHS:
                history_path = user_details.home_path.joinpath(ps_path)
                for history_file in history_path.glob("*_history.txt"):
                    self._history.append((user_details.user, history_file))

    def check_compatible(self) -> None:
        if not self._history:
            raise UnsupportedPluginError("No ConsoleHost_history.txt files found")

    @export(record=ConsoleHostHistoryRecord)
    def powershell_history(self) -> Iterator[ConsoleHostHistoryRecord]:
        """Return PowerShell command history for all users.

        The PowerShell ``ConsoleHost_history.txt`` file contains information about the commands executed with PowerShell in
        a terminal. No data is recorded from terminal-less PowerShell sessions. Commands are saved to disk after the process has completed.
        PSReadLine does not save commands containing 'password', 'asplaintext', 'token', 'apikey' or 'secret'.

        References:
            - https://0xdf.gitlab.io/2018/11/08/powershell-history-file.html
            - https://learn.microsoft.com/en-us/powershell/module/microsoft.powershell.core/about/about_history?view=powershell-7.3#order-of-commands-in-the-history
            - https://learn.microsoft.com/en-us/powershell/module/psreadline/about/about_psreadline?view=powershell-7.3#command-history
        """  # noqa E501

        for user, path in self._history:
            file_mtime = path.stat().st_mtime

            i = 0
            for line in path.open("r"):
                line = line.strip()
                if not line:
                    continue

                yield ConsoleHostHistoryRecord(
                    mtime=file_mtime,
                    command=line,
                    order=i,
                    source=path,
                    _target=self.target,
                    _user=user,
                )

                i += 1



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dissect.target.helpers.record import UnixUserRecord
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix.bsd._os import BsdPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target

RE_CONFIG_IP = re.compile(r"-IPAddress (?P<ip>[^ ]+) ")
RE_CONFIG_HOSTNAME = re.compile(r"set ns hostName (?P<hostname>[^\n]+)\n")
RE_CONFIG_TIMEZONE = re.compile(
    r'set ns param.* -timezone "GMT\+(?P<hours>[0-9]+):(?P<minutes>[0-9]+)-.*-(?P<zone_name>.+)"'
)
RE_CONFIG_USER = re.compile(r"bind system user (?P<user>[^ ]+) ")
RE_LOADER_CONFIG_KERNEL_VERSION = re.compile(r'kernel="/(?P<version>.*)"')


class CitrixPlugin(BsdPlugin):
    def __init__(self, target: Target):
        super().__init__(target)
        self._ips = []
        self._hostname = None
        self._config_usernames = []
        self._parse_netscaler_configs()

    def _parse_netscaler_configs(self) -> None:
        ips = set()
        usernames = set()
        for config_path in self.target.fs.path("/flash/nsconfig/").glob("ns.conf*"):
            with config_path.open("rt") as config_file:
                config = config_file.read()
                for match in RE_CONFIG_IP.finditer(config):
                    ips.add(match.groupdict()["ip"])
                for match in RE_CONFIG_USER.finditer(config):
                    usernames.add(match.groupdict()["user"])
                if config_path.name == "ns.conf":
                    # Current configuration of the netscaler
                    if hostname_match := RE_CONFIG_HOSTNAME.search(config):
                        self._hostname = hostname_match.groupdict()["hostname"]
                    if timezone_match := RE_CONFIG_TIMEZONE.search(config):
                        tzinfo = timezone_match.groupdict()
                        self.target.timezone = tzinfo["zone_name"]

        self._config_usernames = list(usernames)
        self._ips = list(ips)

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        ramdisk = None
        for fs in target.filesystems:
            # /netscaler can be present on both the ramdisk and the harddisk. Therefore we also check for the /log
            # folder, which is not present on the ramdisk. We regard the harddisk as the system volume, as it is
            # possible to only have a disk image of a Netscaler. However, in the case where we only have the ramdisk,
            # we want to fall back on that as the system volume. Thus we store that filesystem in a fallback variable.
            if fs.exists("/netscaler"):
                if fs.exists("/log"):
                    return fs
                ramdisk = fs

        # At this point, we could not find the filesystem for '/var'. Thus, we fall back to the ramdisk variable, which
        # is either 'None' (in which case this isn't a Citrix netscaler), or points to the filesystem of the ramdisk.
        return ramdisk

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        # A disk image of a Citrix Netscaler contains two partitions, that after boot are mounted to /var and /flash.
        # The rest of the filesystem is recreated at runtime into a 'ramdisk'. Currently, this plugin does not
        # yet support recreating the ramdisk from a 'clean' state. This might be possible in a future iteration but
        # requires further research.

        # When the ramdisk is present within the target's filesystems, mount it accordingly,
        for fs in target.filesystems:
            if fs.exists("/bin/freebsd-version"):
                # If available, mount the ramdisk first.
                target.fs.mount("/", fs)
        # The 'disk' filesystem is mounted at '/var'.
        target.fs.mount("/var", sysvol)

        # Enumerate filesystems for flash partition
        for fs in target.filesystems:
            if fs.exists("/nsconfig") and fs.exists("/boot"):
                target.fs.mount("/flash", fs)

        return cls(target)

    @export(property=True)
    def hostname(self) -> str | None:
        return self._hostname or super().hostname

    @export(property=True)
    def version(self) -> str | None:
        version = "Unknown"
        version_path = self.target.fs.path("/flash/.version")
        if version_path.is_file():
            version = version_path.read_text().strip()

        loader_conf_path = self.target.fs.path("/flash/boot/loader.conf")
        if loader_conf_path.is_file():
            loader_conf = loader_conf_path.read_text()
            if match := RE_LOADER_CONFIG_KERNEL_VERSION.search(loader_conf):
                kernel_version = match.groupdict()["version"]
                version = f"{version} ({kernel_version})" if version else kernel_version

        if not version:
            self.target.log.warning("Could not determine kernel version")

        return version

    @export(property=True)
    def ips(self) -> list[str]:
        return self._ips

    @export(record=UnixUserRecord)
    def users(self) -> Iterator[UnixUserRecord]:
        nstmp_users = set()
        seen = set()
        nstmp_path = self.target.fs.path("/var/nstmp/")

        # Build a set of nstmp users
        if nstmp_path.exists():
            for entry in nstmp_path.iterdir():
                if entry.is_dir() and entry.name != "#nsinternal#":
                    # The nsmonitor user has a home directory of /var/nstmp/monitors rather than /var/nstmp/nsmonitor
                    username = "nsmonitor" if entry.name == "monitors" else entry.name
                    nstmp_users.add(username)

        # Yield users from the config, matching them to their 'home' in /var/nstmp if it exists.
        for username in self._config_usernames:
            nstmp_home = nstmp_path.joinpath(username)
            user_home = nstmp_home if nstmp_home.exists() else None

            if user_home:
                # After this loop we will yield all users who are not in the config, but are listed in /var/nstmp/
                # To prevent double records, we remove entries from the set that we are already yielding here.
                nstmp_users.remove(username)

            if username == "root" and self.target.fs.exists("/root"):
                # If we got here, 'root' is present both in /var/nstmp and in /root. In such cases, we yield
                # the 'root' user as having '/root' as a home, not in /var/nstmp, as there is no 'nscli_history'
                # for the root user in /var/nstmp.
                user_home = self.target.fs.path("/root")

            seen.add((username, user_home.as_posix() if user_home else None, None))
            yield UnixUserRecord(name=username, home=user_home)

        # Yield all users in nstmp that were not observed in the config
        for username in nstmp_users:
            # The nsmonitor user has a home directory of /var/nstmp/monitors rather than /var/nstmp/nsmonitor
            home = nstmp_path.joinpath(username) if username != "nsmonitor" else nstmp_path.joinpath("monitors")
            seen.add((username, home.as_posix(), None))
            yield UnixUserRecord(name=username, home=home)

        # Yield users from /etc/passwd if we have not seem them in previous loops
        for user in super().users():
            if (user.name, user.home.as_posix(), user.shell) in seen:
                continue
            # To prevent bogus command history for all users without a home whenever a history is located at the root
            # of the filesystem, we set the user home to None if their home is equivalent to '/'
            user.home = user.home if user.home != "/" else None
            yield user

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.CITRIX.value



from __future__ import annotations

import io
import textwrap
from logging import getLogger
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.target.exceptions import ConfigurationParsingError, FileNotFoundError
from dissect.target.filesystem import FilesystemEntry, VirtualFilesystem
from dissect.target.helpers import fsutil
from dissect.target.helpers.configutil import ConfigurationParser, parse

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

log = getLogger(__name__)


class ConfigurationFilesystem(VirtualFilesystem):
    """A special :class:`.Filesystem` class that allows you to browse and interact with configuration files
    as ``directories`` and ``files`` by parsing the file into key/value pairs.

    Depending on the ``value`` of a configuration file's ``key``, it will act like a ``directory`` or ``file``:
        * When the ``key`` contains sub-values (dictionary), it will act like a ``directory``.
        * Otherwise it will act like a ``file``.

    Examples:
        >>> fs = ConfigurationFilesystem(target, "/etc")
        >>> entry = fs.get("xattr.conf")
        <ConfigurationEntry path=/etc/xattr.conf value=<dissect.target.helpers.configutil.Default object at 0x115683280>
        >>> entry.listdir()  # listed entries are the keys in the configuration file
        [...
        'system.posix_acl_access',
        'system.posix_acl_default',
        'trusted.SGI_ACL_DEFAULT',
        'trusted.SGI_ACL_FILE',
        ...]
        >>> entry.get("system.posix_acl_access")  # returns the value of the key
        <ConfigurationEntry path=/etc/xattr.conf/system.posix_acl_access value=permissions>
        >>> entry.get("system.posix_acl_access").open().read()  # returns the raw value of the key
        b'permissions\\n'

    """

    __type__: str = "META:configuration"

    def __init__(self, target: Target, path: str, **kwargs):
        super().__init__(**kwargs)
        self.root.top = target.fs.get(path)

    def _get_till_file(self, path: str, relentry: FilesystemEntry | None) -> tuple[list[str], FilesystemEntry]:
        """Searches for the file entry that is pointed to by ``path``.

        The ``path`` could contain ``key`` entries too, so it searches for the entry from
        the start of the path.

        Returns:
            A list of ``parts`` containing keys: [keys, into, the, file].
            And the resolved entry: Entry(filename)
        """

        entry = relentry or self.root
        root_path = relentry.path if relentry else self.root.top.path

        # Calculate the relative path
        relpath = fsutil.relpath(path, root_path, alt_separator=self.alt_separator)
        path = fsutil.normalize(relpath, alt_separator=self.alt_separator).strip("/")

        if not path:
            return [], entry

        parts = path.split("/")

        for idx, part in enumerate(parts, start=1):  # noqa: B007
            # Resolve link
            if entry.is_symlink():
                entry = entry.readlink_ext()

            if part == ".":
                continue
            if part == "..":
                entry = entry.up or self.root
                continue

            if entry is self.root:
                entry = self.root.top

            entry = entry.get(part)
            if entry.is_file():
                break

        return parts[idx:], entry

    def get(self, path: str, relentry: FilesystemEntry | None = None, *args, **kwargs) -> ConfigurationEntry:
        """Retrieve a :class:`ConfigurationEntry` relative to the root or ``relentry``.

        Raises:
            FileNotFoundError: if it could not find the entry.
        """
        parts, entry = self._get_till_file(path, relentry)

        if entry.is_dir():
            return ConfigurationEntry(self, entry.path, entry, None)

        entry = self._convert_entry(entry, *args, **kwargs)

        for part in parts:
            _prev_entry = entry
            entry = entry.get(part)
            if entry is None:
                raise FileNotFoundError(f"{part!r} not found in {_prev_entry}.")

        return entry

    def _convert_entry(self, file_entry: FilesystemEntry, *args, **kwargs) -> ConfigurationEntry:
        """Creates a :class:`ConfigurationEntry` from a ``file_entry``.

        If an error occurs during the parsing of the file contents, the original ``file_entry`` is returned.
        """
        entry = file_entry
        config_parser = None
        try:
            target_path = entry.fs.path(entry.path)
            config_parser = parse(target_path, *args, **kwargs)
        except ConfigurationParsingError as e:
            # If a parsing error gets created, it should return the `entry`
            log.debug("Error when parsing %s with message '%s'", entry.path, e)

        return ConfigurationEntry(self, entry.path, entry, config_parser)


class ConfigurationEntry(FilesystemEntry):
    """A Special filesystem entry.

    Behaves like a ``directory`` when :attr:`parser_items` is a :class:`.ConfigurationParser` or a ``dict``.
    Behaves like a ``file`` otherwise.

    Args:
        parser_items: A dict-like object containing all configuration entries and values.
            In most cases this is either a :class:`.ConfigurationParser` or ``dict``.
            Otherwise, its the entry's value

    Examples:
        >>> fs = ConfigurationFilesystem(target, "/etc")
        >>> entry = fs.get("xattr.conf")
        <ConfigurationEntry path=/etc/xattr.conf value=<dissect.target.helpers.configutil.Default object at 0x115683280>
        >>> entry.listdir()  # listed entries are the keys in the configuration file
        [...
        'system.posix_acl_access',
        'system.posix_acl_default',
        'trusted.SGI_ACL_DEFAULT',
        'trusted.SGI_ACL_FILE',
        ...]
        >>> entry.as_dict()
        {
            ...
            "system.posix_acl_access" : "...",
            ...
        }
        >>> entry.get("system.posix_acl_access")  # returns the value of the key
        <ConfigurationEntry path=/etc/xattr.conf/system.posix_acl_access value=permissions>
        >>> entry.get("system.posix_acl_access").open().read()  # returns the raw value of the key
        b'permissions\\n'
    """

    def __init__(
        self,
        fs: ConfigurationFilesystem,
        path: str,
        entry: FilesystemEntry,
        parser_items: dict | ConfigurationParser | str | list | None = None,
    ) -> None:
        super().__init__(fs, path, entry)
        self.parser_items = parser_items

    def __getitem__(self, item: str) -> ConfigurationEntry:
        return self.parser_items[item]

    def __repr__(self) -> str:
        output = f"path={self.path}"

        if not isinstance(self.parser_items, dict):
            output += f" value={self.parser_items}"

        return f"<{self.__class__.__name__} {output}"

    def get(self, key: str, default: Any | None = None) -> ConfigurationEntry | Any | None:
        """Gets the dictionary key that belongs to this entry using ``key``.
        Behaves like ``dictionary.get()``.

        Args:
            ``key``: A dictionary key that is inside :attr:`parser_items`.
            ``default``: The default value to return if ``key`` is not inside this entry.

        Returns:
            a :class:`ConfigurationEntry` when ``key`` is present, otherwise its ``default``.
        """
        # Check for path in config entry
        if not key:
            raise ValueError("key should be defined")

        path = fsutil.join(self.path, key, alt_separator=self.fs.alt_separator)

        if self.parser_items and key in self.parser_items:
            return ConfigurationEntry(
                self.fs,
                path,
                self.entry,
                self.parser_items[key],
            )

        if self.entry.is_dir():
            return self.fs.get(path, self.entry)

        return default

    def _write_value_mapping(self, values: dict[str, Any], indentation_nr: int = 0) -> str:
        """Internal function to transform the ``values`` dictionary to a string representation.

        Args:
            values: A dictionary or its contents.
            indentation_nr: How much indentation this entry should receive.

        Returns:
            An indented string containing all the information inside ``values``.
        """
        prefix = " " * indentation_nr
        output_buffer = io.StringIO()

        if isinstance(values, list):
            # Explicitly convert the list values to strings
            _text = "\n".join(str(val) for val in values)
            output_buffer.write(textwrap.indent(text=_text, prefix=prefix))
        elif hasattr(values, "keys"):
            for key, value in values.items():
                output_buffer.write(textwrap.indent(key, prefix=prefix) + "\n")
                output_buffer.write(self._write_value_mapping(value, indentation_nr + 4))
        else:
            output_buffer.write(textwrap.indent(str(values), prefix=prefix) + "\n")

        return output_buffer.getvalue()

    def open(self) -> BinaryIO:
        """Open this :class:`ConfigurationEntry`.

        If :attr:`parser_items` is a :class:`.ConfigurationParser`,
        it will ``open`` the underlying entry.

        Returns:
            A file-like object holding a byte representation of :attr:`parser_items`.
        """

        if isinstance(self.parser_items, ConfigurationParser):
            # Currently trying to open the underlying entry
            return self.entry.open()

        if isinstance(self.parser_items, bytes):
            return io.BytesIO(self.parser_items)

        output_data = self._write_value_mapping(self.parser_items)
        return io.BytesIO(bytes(output_data, "utf-8"))

    def iterdir(self) -> Iterator[str]:
        for entry in self.scandir():
            yield entry.name

    def scandir(self) -> Iterator[ConfigurationEntry]:
        """Return the items inside :attr:`parser_items` as ``ConfigurationEntries``."""
        if self.is_file():
            raise NotADirectoryError

        if self.parser_items is None and self.entry.is_dir():
            for entry in self.entry.scandir():
                yield ConfigurationEntry(self.fs, entry.name, entry, None)
            return

        for key, values in self.parser_items.items():
            yield ConfigurationEntry(self.fs, key, self.entry, values)

    def is_file(self, follow_symlinks: bool = True) -> bool:
        return not self.is_dir(follow_symlinks)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Returns whether this :class:`ConfigurationEntry` can be considered a directory."""
        # if self.parser_items has keys (thus sub-values), we can consider it a directory.
        return (self.parser_items is None and self.entry.is_dir()) or hasattr(self.parser_items, "keys")

    def is_symlink(self) -> bool:
        """Return whether this :class:`ConfigurationEntry` is a symlink or not.

        Returns:
            False, as ``ConfigurationEntries`` are never symlinks.
        """
        # ConfigurationEntries are already resolved, so are never symlinks.
        return False

    def exists(self, key: str) -> bool:
        """Return whether the underlying :class:`.FilesystemEntry` :attr:`entry` and
        supplied ``key`` exists inside this :class:`ConfigurationEntry`.

        Returns:
            Whether the ``entry`` and ``key`` exists
        """
        return self.parser_items and self.entry.exists() and key in self.parser_items

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Returns the stat from the underlying :class:`.FilesystemEntry` :attr:`entry`."""
        return self.entry.stat(follow_symlinks)

    def lstat(self) -> fsutil.stat_result:
        """Returns the lstat from the underlying :class:`.FilesystemEntry` :attr:`entry`."""

        return self.entry.lstat()

    def as_dict(self) -> dict:
        """Returns :attr:`parser_items` as a dictionary."""
        if isinstance(self.parser_items, ConfigurationParser):
            return self.parser_items.parsed_data
        return self.parser_items



from __future__ import annotations

import datetime
import io
import stat
import textwrap
from unittest.mock import mock_open, patch

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.helpers import fsutil, utils


def test_to_list_single_value() -> None:
    assert utils.to_list(1) == [1]
    assert utils.to_list("a") == ["a"]
    assert utils.to_list(None) == []


def test_to_list_list_value() -> None:
    assert utils.to_list([1, 2, 3]) == [1, 2, 3]
    assert utils.to_list(["a", "b", "c"]) == ["a", "b", "c"]
    assert utils.to_list([]) == []


def test_slugify() -> None:
    assert utils.slugify("foo/bar\\baz bla") == "foo_bar_baz_bla"


def test_filesystem_readinto() -> None:
    data = b"hello_world"
    mocked_file = mock_open(read_data=b"hello_world")

    buffer = bytearray([0] * 512)
    assert utils.readinto(buffer, mocked_file.return_value) == len(data)
    assert buffer[: len(data)] == data
    assert len(buffer) == 512


def test_helpers_fsutil_year_rollover_helper() -> None:
    vfs = VirtualFilesystem()

    content = """
    Dec 31 03:14:15 Line 1
    Jan  1 13:21:34 Line 2
    Dec 31 03:14:15 Line 3
    Jan  1 13:21:34 Line 4
    Dec 31 03:14:15 Line 5
    Jan  1 13:21:34 Line 6
    Jan  2 13:21:34 Line 7
    Feb  3 13:21:34 Line 8
    Dec 31 03:14:15 Line 9
    Jan  1 13:21:34 Line 10
    """
    re_ts = r"(\w+\s{1,2}\d+\s\d{2}:\d{2}:\d{2})"
    ts_fmt = "%b %d %H:%M:%S"

    vfs.map_file_fh("file", io.BytesIO(textwrap.dedent(content).encode()))
    path = vfs.path("file")

    mocked_stat = fsutil.stat_result([stat.S_IFREG, 1337, id(vfs), 0, 0, 0, len(content), 0, 3384460800, 0])
    with patch.object(path, "stat", return_value=mocked_stat):
        result = list(utils.year_rollover_helper(path, re_ts, ts_fmt))
        assert len(result) == 10

        year_line = [(ts.year, line) for ts, line in result]

        assert result[0][0].tzinfo == datetime.timezone.utc

        assert year_line[0] == (2077, "Jan  1 13:21:34 Line 10")
        assert year_line[-1] == (2073, "Dec 31 03:14:15 Line 1")

    # This mtime is in 2023 in UTC, but 2022 in -8
    # Test that the year detection correctly starts at 2022 local time if we parse with a timezone of -8
    mtime = datetime.datetime(2023, 1, 1, 3, tzinfo=datetime.timezone.utc)
    tzinfo = datetime.timezone(datetime.timedelta(hours=-8))
    mocked_stat = fsutil.stat_result([stat.S_IFREG, 1337, id(vfs), 0, 0, 0, len(content), 0, mtime.timestamp(), 0])
    with patch.object(path, "stat", return_value=mocked_stat):
        result = list(utils.year_rollover_helper(path, re_ts, ts_fmt, tzinfo))
        year_line = [(ts.year, line) for ts, line in result]

        assert result[0][0].tzinfo == tzinfo

        assert year_line[0] == (2022, "Jan  1 13:21:34 Line 10")
        assert year_line[-1] == (2018, "Dec 31 03:14:15 Line 1")


def test_helpers_fsutil_year_rollover_helper_leap_day() -> None:
    """Test if we correctly handle leap days such as 2024-02-29."""

    content = """
    Feb 28 11:00:00 Line 1
    Feb 29 12:00:00 Line 2
    Mar  1 13:00:00 Line 3
    """
    fs = VirtualFilesystem()
    fs.map_file_fh("file", io.BytesIO(textwrap.dedent(content).encode()))
    path = fs.path("file")

    re_ts = r"(\w+\s{1,2}\d+\s\d{2}:\d{2}:\d{2})"
    ts_fmt = "%b %d %H:%M:%S"
    mocked_stat = fsutil.stat_result(
        [stat.S_IFREG, 1337, id(fs), 0, 0, 0, len(content), 0, 1709294400, 0]
    )  # mtime is set to 2024-03-01.

    with patch.object(path, "stat", return_value=mocked_stat):
        result = list(utils.year_rollover_helper(path, re_ts, ts_fmt))
        assert len(result) == 3



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

import dissect.target.plugins.os.windows.prefetch as prefetch

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.cstruct import cstruct


@pytest.fixture
def mocked_cstruct(version: int) -> Iterator[cstruct]:
    with patch.object(prefetch, "c_prefetch") as mocked_cstruct:
        mocked_cstruct.PREFETCH_HEADER.return_value.version = version
        yield mocked_cstruct


@pytest.fixture
def mocked_prefetch() -> prefetch.Prefetch:
    with patch.object(prefetch, "c_prefetch"), patch.multiple(prefetch.Prefetch, identify=Mock(), parse=Mock()):
        return prefetch.Prefetch(Mock())


@pytest.mark.parametrize(
    ("version", "dict_output"),
    [
        (17, ("FILE_INFORMATION_17", "FILE_METRICS_ARRAY_ENTRY_17")),
        (23, ("FILE_INFORMATION_23", "FILE_METRICS_ARRAY_ENTRY_23")),
        (30, ("FILE_INFORMATION_26", "FILE_METRICS_ARRAY_ENTRY_23")),
        (31, ("FILE_INFORMATION_26", "FILE_METRICS_ARRAY_ENTRY_23")),
    ],
)
def test_prefetch_valid_versions(mocked_cstruct: cstruct, version: int, dict_output: tuple[str, str]) -> None:
    file_info = getattr(mocked_cstruct, dict_output[0])
    metric_array = getattr(mocked_cstruct, dict_output[1])
    with (
        patch.dict(prefetch.prefetch_version_structs, {version: (file_info, metric_array)}),
        patch.object(prefetch.Prefetch, "parse_metrics") as mocked_metric,
    ):
        prefetch_obj = prefetch.Prefetch(Mock())
        assert prefetch_obj.fn == file_info.return_value
        mocked_metric.assert_called_with(metric_array_struct=metric_array)


@pytest.mark.parametrize("version", [0xDEADBEEF])
def test_prefetch_invalid_version(mocked_cstruct: cstruct) -> None:
    with pytest.raises(NotImplementedError):
        prefetch.Prefetch(Mock())


def test_prefetch_datetime(mocked_prefetch: prefetch.Prefetch) -> None:
    mocked_prefetch.fn = Mock()
    mocked_prefetch.fn.last_run_time = 0xDEADBEEF

    with patch("dissect.target.plugins.os.windows.prefetch.wintimestamp") as wintimestamp:
        assert mocked_prefetch.latest_timestamp == wintimestamp.return_value


def test_prefetch_unknown_attribute(mocked_prefetch: prefetch.Prefetch) -> None:
    mocked_prefetch.fn = Mock(spec=[])
    assert len(mocked_prefetch.previous_timestamps) == 0


@pytest.mark.parametrize(
    ("dates", "expected_length"),
    [
        ([0x0] * 12, 0),
        ([0x1] * 8, 8),
        ([0x1, 0x2, 0x0, 0x1], 3),
        ([0x1, 0x0, 0x0, 0x20, 0x0], 2),
    ],
)
def test_prefetch_last_run_dates(mocked_prefetch: prefetch.Prefetch, dates: list[int], expected_length: int) -> None:
    mocked_prefetch.fn = Mock()
    mocked_prefetch.fn.last_run_remains = dates

    assert len(mocked_prefetch.previous_timestamps) == expected_length


def test_prefetch_parse_metrics(mocked_prefetch: prefetch.Prefetch) -> None:
    mocked_prefetch.fh = Mock()
    mocked_prefetch.fn = Mock()
    mocked_prefetch.fn.number_of_file_metrics_entries = 10
    with patch.object(prefetch.Prefetch, "read_filename") as filename:
        filename.return_value.decode.return_value = ""
        metric_struct = Mock()
        mocked_prefetch.fn.filename_strings_offset = ""
        metric_struct.return_value.filename_string_offset = ""
        test = mocked_prefetch.parse_metrics(metric_array_struct=metric_struct)
        assert len(test) == 10


def test_prefetch_read_filename(mocked_prefetch: prefetch.Prefetch) -> None:
    mocked_fileheader = Mock()
    mocked_fileheader.read.return_value = b""

    mocked_prefetch.fh = mocked_fileheader
    filename = mocked_prefetch.read_filename(0x10, 0x10)

    mocked_fileheader.read.assert_called_with(0x10 * 2)
    assert isinstance(filename, bytes)
    assert mocked_fileheader.seek.call_count == 2



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock, mock_open, patch

import pytest
from flow.record.fieldtypes import command, digest, path

from dissect.target.exceptions import FileNotFoundError, IsADirectoryError
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.record_modifier import (
    Modifier,
    ModifierFunc,
    get_modifier_function,
)
from tests.helpers.test_hashutil import HASHES

if TYPE_CHECKING:
    from flow.record import Record

    from dissect.target.target import Target


@pytest.fixture
def hash_function() -> ModifierFunc:
    return get_modifier_function(Modifier.HASH)


@pytest.fixture
def resolve_function() -> ModifierFunc:
    return get_modifier_function(Modifier.RESOLVE)


@pytest.mark.parametrize(
    ("test_input", "expected_records"),
    [
        ({"name": path}, 2),
        ({"name": path, "test": path}, 3),
        ({"name": path, "test": str}, 2),
        ({"name": command}, 2),
    ],
)
@patch("flow.record.Record")
def test_hash_path_records_with_paths(
    record: Record,
    hash_function: ModifierFunc,
    target_win: Mock,
    test_input: dict[str, type[path | str]],
    expected_records: int,
) -> None:
    record._desc.name = "test"
    record._field_types = test_input

    path_field_names = [key for key, value in test_input.items() if value is path]

    with (
        patch.object(TargetPath, "open", mock_open(read_data=b"")),
        patch("dissect.target.helpers.fsutil.TargetPath.exists", return_value=True),
        patch("dissect.target.helpers.fsutil.TargetPath.is_file", return_value=True),
        patch("dissect.target.helpers.record_modifier.common", return_value=HASHES),
    ):
        hashed_record = hash_function(target_win, record)

    assert hashed_record.name == "test"
    assert len(hashed_record.records) == expected_records
    assert hashed_record.records[0] == record

    for name, _record in zip(path_field_names, hashed_record.records[1:]):
        assert getattr(_record, f"{name}_resolved") is not None
        assert getattr(_record, f"{name}_digest").__dict__ == digest(HASHES).__dict__


@pytest.mark.parametrize(
    "test_input",
    [
        {},
        {"name": str},
        {"name": str, "test": str},
    ],
)
@patch("flow.record.Record")
def test_hash_path_records_without_paths(
    record: Record,
    hash_function: ModifierFunc,
    test_input: dict[str, type[str]],
) -> None:
    record._desc.name = "test"
    record._field_types = test_input

    assert hash_function(Mock(), record) == record


@pytest.mark.parametrize(
    ("side_effects", "expected"),
    [
        ([FileNotFoundError], 0),
        ([IsADirectoryError], 0),
        ([FileNotFoundError, FileNotFoundError], 0),
        ([IsADirectoryError, IsADirectoryError], 0),
        ([FileNotFoundError, IsADirectoryError], 0),
        ([FileNotFoundError, HASHES], 1),
        ([IsADirectoryError, HASHES], 1),
        ([FileNotFoundError, HASHES, IsADirectoryError], 1),
        ([HASHES, FileNotFoundError, HASHES], 2),
        ([HASHES, IsADirectoryError, HASHES], 2),
    ],
)
@patch("flow.record.Record")
def test_hash_path_records_with_exception(
    record: Record,
    hash_function: ModifierFunc,
    target_win: Target,
    side_effects: list[type[Exception] | tuple[str]],
    expected: int,
) -> None:
    record._desc.name = "test"
    field_types = {}
    found_type_names = []
    for idx, data in enumerate(side_effects):
        field_name = f"path_{idx}"
        field_types[field_name] = path
        setattr(record, field_name, "test")
        if data is HASHES:
            found_type_names.append(field_name)

    record._field_types = field_types

    mocked_open = mock_open()
    mocked_open.configure_mock(**{"return_value.__enter__.side_effect": side_effects})

    with (
        patch.object(TargetPath, "open", mocked_open),
        patch("dissect.target.helpers.record_modifier.common", return_value=HASHES),
    ):
        hashed_record = hash_function(target_win, record)

    if not expected:
        assert hashed_record == record
    else:
        for _record, key in zip(hashed_record.records[1:], found_type_names):
            assert getattr(_record, f"{key}_resolved") == "test"
            assert getattr(_record, f"{key}_digest").__dict__ == digest(HASHES).__dict__


@patch("flow.record.Record")
def test_resolved_modifier(record: Record, target_win: Target, resolve_function: ModifierFunc) -> None:
    record._desc.name = "test"
    record._field_types = {"name": path}

    resolved_record = resolve_function(target_win, record)

    for _record in resolved_record.records[1:]:
        assert _record.name_resolved is not None
        assert not hasattr(_record, "name_digest")



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from unittest.mock import Mock

import pytest
from flow.record.fieldtypes import windows_path

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.plugins.os.windows.generic import ComputerSidRecord
from dissect.target.plugins.os.windows.registry import RegistryPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


def current_version_key() -> str:
    hive_name, path = WindowsPlugin.CURRENT_VERSION_KEY.split("\\", maxsplit=1)
    hive_name = RegistryPlugin.SHORTNAMES.get(hive_name, hive_name)
    return f"{hive_name}\\{path}"


CURRENT_VERSION_KEY = current_version_key()


@pytest.fixture
def version_target(target_win: Target) -> Target:
    hive = target_win.registry._root
    hive.map_key(CURRENT_VERSION_KEY, VirtualKey(hive, CURRENT_VERSION_KEY))
    return target_win


@pytest.fixture
def win_plugin(version_target: Target) -> WindowsPlugin:
    return WindowsPlugin(version_target)


@pytest.fixture
def target_win_linux_folders(target_win: Filesystem, fs_linux_sys: Filesystem) -> Target:
    target_win.fs.mount("/", fs_linux_sys)
    return target_win


def map_version_value(target: Target, name: str | None, value: Any) -> None:
    if name is not None:
        hive = target.registry._root
        hive.map_value(CURRENT_VERSION_KEY, name, VirtualValue(hive, name, value))


def assert_value(result: Any, value: Any) -> None:
    if value is None:
        assert result is value
    else:
        assert result == value


@pytest.mark.parametrize(
    ("name", "value"),
    [
        (None, None),
        ("CurrentVersion", "Some Stringy Version"),
    ],
)
def test_windowsplugin__legacy_curre_ntversion(
    version_target: Target,
    win_plugin: WindowsPlugin,
    name: str | None,
    value: Any,
) -> None:
    map_version_value(version_target, name, value)
    result = win_plugin._legacy_current_version()

    assert_value(result, value)


@pytest.mark.parametrize(
    ("name", "value"),
    [
        (None, None),
        ("CurrentMajorVersionNumber", 10),
    ],
)
def test_windowsplugin__major_version(
    version_target: Target,
    win_plugin: WindowsPlugin,
    name: str | None,
    value: Any,
) -> None:
    map_version_value(version_target, name, value)
    result = win_plugin._major_version()

    assert_value(result, value)


@pytest.mark.parametrize(
    ("name", "value"),
    [
        (None, None),
        ("CurrentMinorVersionNumber", 0),
    ],
)
def test_windowsplugin__minor_version(
    version_target: Target,
    win_plugin: WindowsPlugin,
    name: str | None,
    value: Any,
) -> None:
    map_version_value(version_target, name, value)
    result = win_plugin._minor_version()

    assert_value(result, value)


@pytest.mark.parametrize(
    ("keys", "value"),
    [
        ([], None),
        ([("CurrentVersion", "x.y")], "x.y"),
        (
            [
                ("CurrentMajorVersionNumber", 10),
                ("CurrentMinorVersionNumber", 0),
            ],
            "10.0",
        ),
        (
            [
                ("CurrentMajorVersionNumber", 10),
                ("CurrentMinorVersionNumber", 0),
                ("CurrentVersion", "x.y"),
            ],
            "10.0",
        ),
        ([("CurrentMajorVersionNumber", 10)], "10."),
        ([("CurrentMinorVersionNumber", 0)], None),
    ],
)
def test_windowsplugin__nt_version(
    version_target: Target,
    win_plugin: WindowsPlugin,
    keys: list[tuple[str, Any]],
    value: str | None,
) -> None:
    for key_name, key_value in keys:
        map_version_value(version_target, key_name, key_value)
    result = win_plugin._nt_version()

    assert_value(result, value)


@pytest.mark.parametrize(
    ("keys", "value"),
    [
        ([], None),
        (
            [
                ("CSDVersion", 5678),
                ("CurrentBuildNumber", 1234),
                ("CurrentMajorVersionNumber", 10),
                ("CurrentMinorVersionNumber", 0),
                ("CurrentVersion", "x.y"),
                ("ProductName", "Some Product"),
                ("UBR", 9012),
            ],
            "Some Product (NT 10.0) 1234.9012 5678",
        ),
        (
            [
                ("CSDVersion", 5678),
                ("CurrentBuildNumber", 1234),
                ("CurrentVersion", "x.y"),
                ("ProductName", "Some Product"),
                ("UBR", 9012),
            ],
            "Some Product (NT x.y) 1234.9012 5678",
        ),
        (
            [
                ("CurrentBuildNumber", 1234),
                ("CurrentVersion", "x.y"),
                ("ProductName", "Some Product"),
            ],
            "Some Product (NT x.y) 1234",
        ),
        (
            [
                ("ProductName", "Some Product"),
            ],
            "Some Product (NT <Unknown CurrentVersion>) <Unknown CurrentBuildNumber>",
        ),
        (
            [
                ("CurrentVersion", "x.y"),
            ],
            "<Unknown ProductName> (NT x.y) <Unknown CurrentBuildNumber>",
        ),
        (
            [
                ("CurrentBuildNumber", 1234),
            ],
            "<Unknown ProductName> (NT <Unknown CurrentVersion>) 1234",
        ),
        (
            [
                ("UBR", 9012),
            ],
            "<Unknown ProductName> (NT <Unknown CurrentVersion>) <Unknown CurrentBuildNumber>.9012",
        ),
        (
            [
                ("CSDVersion", 5678),
            ],
            "<Unknown ProductName> (NT <Unknown CurrentVersion>) <Unknown CurrentBuildNumber> 5678",
        ),
        (
            [
                ("ProductName", "Windows 10 Pro"),
                ("CurrentMajorVersionNumber", 10),
                ("CurrentMinorVersionNumber", 0),
                ("CurrentBuildNumber", 19_045),
                ("UBR", 1234),
            ],
            "Windows 10 Pro (NT 10.0) 19045.1234",
        ),
        (
            [
                ("ProductName", "Windows 10 Enterprise"),
                ("CurrentMajorVersionNumber", 10),
                ("CurrentMinorVersionNumber", 0),
                ("CurrentBuildNumber", 22_000),
                ("UBR", 1234),
            ],
            "Windows 11 Enterprise (NT 10.0) 22000.1234",
        ),
    ],
)
def test_windowsplugin_version(
    version_target: Target,
    win_plugin: WindowsPlugin,
    keys: list[tuple[str, Any]],
    value: str | None,
) -> None:
    for key_name, key_value in keys:
        map_version_value(version_target, key_name, key_value)
    result = win_plugin.version

    assert_value(result, value)


def test_windows_os_detection_with_linux_folders(target_win_linux_folders: Target) -> None:
    fs_linux = LinuxPlugin.detect(target_win_linux_folders)
    fs_windows = WindowsPlugin.detect(target_win_linux_folders)

    assert fs_linux is None
    assert fs_windows is not None


def test_windows_user(target_win_users: Target) -> None:
    users = list(target_win_users.users())

    assert len(users) == 2

    assert users[0].sid == "S-1-5-18"
    assert users[0].name == "systemprofile"
    assert users[0].home == windows_path("c:\\Windows\\system32\\config\\systemprofile")

    assert users[1].sid == "S-1-5-21-3263113198-3007035898-945866154-1002"
    assert users[1].name == "John"
    assert users[1].home == windows_path("C:\\Users\\John")


def test_windows_user_from_sam(target_win_users: Target) -> None:
    """Verify the home folder name is independent of the SAM username.

    The final component of the home folder path (e.g., 'C:\\Users\\Username')
    is not always identical to the account's username after a rename. This
    test ensures our code correctly handles this distinction.

    References:
        - https://learn.microsoft.com/en-us/previous-versions/troubleshoot/windows-client/renaming-user-account-not-change-profile-path
    """

    fake_sam_user = Mock()
    fake_sam_user.rid = 1002
    fake_sam_user.username = "Jane"  # Should override "John" from home folder

    target_win_users.machine_sid = Mock(
        return_value=iter([ComputerSidRecord(sid="S-1-5-21-3263113198-3007035898-945866154")])
    )
    target_win_users.sam = Mock(return_value=[fake_sam_user])

    users = list(target_win_users.users())

    # There should be two users, SYSTEM and the user from SAM
    assert len(users) == 2

    # SYSTEM user
    assert users[0].sid == "S-1-5-18"
    assert users[0].name == "systemprofile"
    assert users[0].home == windows_path("c:\\Windows\\system32\\config\\systemprofile")

    # User from SAM, username from SAM should have priority
    assert users[1].sid == "S-1-5-21-3263113198-3007035898-945866154-1002"
    assert users[1].name == "Jane"
    assert users[1].home == windows_path("C:\\Users\\John")


@pytest.mark.parametrize(
    ("registry_value", "expected_hostname"),
    [
        (b"DESKTOP-EXAMPLE", "DESKTOP-EXAMPLE"),
    ],
)
def test_windows_hostname(
    registry_value: bytes, expected_hostname: str, target_win_users: Target, hive_hklm: VirtualHive
) -> None:
    """Test if we can parse windows hostnames correctly."""

    key_name = "SYSTEM\\ControlSet001\\Control\\ComputerName\\ComputerName"
    key = VirtualKey(hive_hklm, key_name)
    key.add_value("ComputerName", VirtualValue(hive_hklm, "ComputerName", registry_value.decode()))
    hive_hklm.map_key(key_name, key)

    assert target_win_users.hostname == expected_hostname
